import os
import random
import string

import environ

env = environ.Env()

DEBUG = False

ANONYMOUS_USER_NAME = "AnonymousUser"

AUTH_USER_MODEL = "testapp.CustomUser"
GUARDIAN_MONKEY_PATCH_USER = False

INSTALLED_APPS = (
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.admin",
    "django.contrib.messages",
    "guardian",
    "guardian.testapp",
)

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    "guardian.backends.ObjectPermissionBackend",
)

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)

TEST_RUNNER = "django.test.runner.DiscoverRunner"

ROOT_URLCONF = "guardian.testapp.tests.urls"
SITE_ID = 1

SECRET_KEY = "".join(random.choice(string.ascii_letters) for x in range(40))

# Database specific

DATABASES = {"default": env.db(default="sqlite:///")}


TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": (os.path.join(os.path.dirname(__file__), "tests", "templates"),),
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
        },
    },
]

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Timezone settings
USE_TZ = True
TIME_ZONE = "UTC"



# Generated by Django 1.9.9 on 2016-09-18 17:52
import datetime

from django.conf import settings
import django.contrib.auth.models
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone

import guardian.mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="CustomUser",
            fields=[
                ("password", models.CharField(max_length=128, verbose_name="password")),
                ("last_login", models.DateTimeField(blank=True, null=True, verbose_name="last login")),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={"unique": "A user with that username already exists."},
                        help_text="Required. 30 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=30,
                        unique=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                "^[\\w.@+-]+$",
                                "Enter a valid username. This value may contain only letters, numbers and @/./+/-/_ characters.",
                            )
                        ],
                        verbose_name="username",
                    ),
                ),
                ("first_name", models.CharField(blank=True, max_length=30, verbose_name="first name")),
                ("last_name", models.CharField(blank=True, max_length=30, verbose_name="last name")),
                ("email", models.EmailField(blank=True, max_length=254, verbose_name="email address")),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                ("date_joined", models.DateTimeField(default=django.utils.timezone.now, verbose_name="date joined")),
                ("custom_id", models.AutoField(primary_key=True, serialize=False)),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "verbose_name": "user",
                "verbose_name_plural": "users",
            },
            bases=(models.Model, guardian.mixins.GuardianUserMixin),
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="CustomUsernameUser",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                ("last_login", models.DateTimeField(blank=True, null=True, verbose_name="last login")),
                ("email", models.EmailField(max_length=100, unique=True)),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model, guardian.mixins.GuardianUserMixin),
        ),
        migrations.CreateModel(
            name="Mixed",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="MixedGroupObjectPermission",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("content_object", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="testapp.Mixed")),
                ("group", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Group")),
                ("permission", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Permission")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="CharPKModel",
            fields=[
                ("char_pk", models.CharField(max_length=128, primary_key=True, serialize=False)),
            ],
        ),
        migrations.CreateModel(
            name="Post",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=64, verbose_name="title")),
            ],
        ),
        migrations.CreateModel(
            name="Project",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, unique=True)),
                ("created_at", models.DateTimeField(default=datetime.datetime.now)),
            ],
            options={
                "get_latest_by": "created_at",
            },
        ),
        migrations.CreateModel(
            name="ProjectGroupObjectPermission",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "content_object",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="testapp.Project"),
                ),
                ("group", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Group")),
                ("permission", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Permission")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ProjectUserObjectPermission",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "content_object",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="testapp.Project"),
                ),
                ("permission", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Permission")),
                ("user", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ReverseMixed",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=128, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="ReverseMixedUserObjectPermission",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "content_object",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="testapp.ReverseMixed"),
                ),
                ("permission", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.Permission")),
                ("user", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AlterUniqueTogether(
            name="reversemixeduserobjectpermission",
            unique_together={("user", "permission", "content_object")},
        ),
        migrations.AlterUniqueTogether(
            name="projectuserobjectpermission",
            unique_together={("user", "permission", "content_object")},
        ),
        migrations.AlterUniqueTogether(
            name="projectgroupobjectpermission",
            unique_together={("group", "permission", "content_object")},
        ),
        migrations.AlterUniqueTogether(
            name="mixedgroupobjectpermission",
            unique_together={("group", "permission", "content_object")},
        ),
    ]



from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from guardian.compat import user_model_label
from guardian.ctypes import get_content_type
from guardian.managers import GroupObjectPermissionManager, UserObjectPermissionManager


class BaseObjectPermission(models.Model):
    """Base ObjectPermission model.

    Child classed should additionally define a `content_object` field and either `user` or `group` field.

    See Also:
        `UserObjectPermission` and `GroupObjectPermission`
    """

    permission = models.ForeignKey(Permission, on_delete=models.CASCADE)

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return "{} | {} | {}".format(
            str(self.content_object), str(getattr(self, "user", False) or self.group), str(self.permission.codename)
        )

    def save(self, *args, **kwargs) -> None:
        """Save the current instance.

        Override this if you need to control the saving process.
        The `force_insert` and `force_update` parameters can be used to insist that the “save”
        must be an SQL insert or update statement, respectively (or equivalent for non-SQL backends).
        Normally, they should not be set.

        Other Parameters:
            force_insert (bool): If True, the save will be forced to be an insert.
            force_update (bool): If True, the save will be forced to be an update.
        """
        content_type = get_content_type(self.content_object)
        if content_type != self.permission.content_type:
            raise ValidationError(
                "Cannot persist permission not designed for "
                "this class (permission's type is %r and object's type is %r)"
                % (self.permission.content_type, content_type)
            )
        return super().save(*args, **kwargs)


class BaseGenericObjectPermission(models.Model):
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_pk = models.CharField(_("object ID"), max_length=255)
    content_object = GenericForeignKey(fk_field="object_pk")

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=["content_type", "object_pk"]),
        ]


class UserObjectPermissionBase(BaseObjectPermission):
    """Base class for creating object level permissions for users.

    This class can be used as a base class for creating a custom model to
    manager user object-level permissions.

    Attributes:
        user (ForeignKey): The django user model that has the permission.
        permission (ForeignKey): Foreign key to the permission granted.
        content_object (ForeignKey, GenericForeignKey):
            A foreign Key to the model class that the permission will be granted for.

    Example:
        ```python
        from guardian.models import UserObjectPermissionBase

        class OrgUserObjectPermission(UserObjectPermissionBase):
            \"\"\"Organization Specific permissions.\"\"\"

            class Meta(UserObjectPermissionBase.Meta):
                verbose_name = "Organization Permission"
                verbose_name_plural = "Organization Permissions"

            # Note: class attribute must be named content_object
            content_object = models.ForeignKey("myapp.Org", on_delete=models.CASCADE, db_column="org_object_id")

        ```

    See Also:
        - [Django-Guardian Performance Tuning](https://django-guardian.readthedocs.io/en/stable/userguide/performance.html)
        - [How to override the default UserObjectPermission](https://django-guardian.readthedocs.io/en/stable/configuration.html#guardian-user-obj-perms-model)
    """

    user = models.ForeignKey(user_model_label, on_delete=models.CASCADE)

    objects = UserObjectPermissionManager()

    class Meta:
        abstract = True
        unique_together = ["user", "permission", "content_object"]


class UserObjectPermissionAbstract(UserObjectPermissionBase, BaseGenericObjectPermission):
    class Meta(UserObjectPermissionBase.Meta, BaseGenericObjectPermission.Meta):
        abstract = True
        unique_together = ["user", "permission", "object_pk"]


class UserObjectPermission(UserObjectPermissionAbstract):
    """The default implementation of the UserObjectPermissionAbstract model.

    If `GUARDIAN_USER_OBJ_PERMS_MODEL` is not set at the beginning of the project, this model will be used.
    Uses Django's contenttypes framework to store generic relations.

    See Also:
        - [Django's Documentation on Abstract Base Models](https://docs.djangoproject.com/en/stable/topics/db/models/#abstract-base-classes)
        - [Django-Guardian Performance Tuning](https://django-guardian.readthedocs.io/en/stable/userguide/performance.html)
        - [How to override the default UserObjectPermission](https://django-guardian.readthedocs.io/en/stable/configuration.html#guardian-user-obj-perms-model)
    """

    class Meta(UserObjectPermissionAbstract.Meta):
        abstract = False
        indexes = [
            models.Index(fields=["permission", "user", "content_type", "object_pk"]),
            models.Index(fields=["user", "content_type", "object_pk"]),
        ]


class GroupObjectPermissionBase(BaseObjectPermission):
    """Base class for creating django-guardian groups.

    This class can be used as a base class for creating a groups permission.

    Attributes:
        group (ForeignKey): A foreign key to the django auth group.
        permission (ForeignKey): Foreign key to the permission granted.
        content_object (ForeignKey, GenericForeignKey):
            A foreign Key to the model class that the permission will be granted for.

    Example:
        ```python
        from guardian.models import GroupObjectPermissionBase

        class OrgGroupObjectPermission(GroupObjectPermissionBase):
            \"\"\"Organization Groups.\"\"\"

            class Meta(GroupObjectPermissionBase.Meta):
                verbose_name = "Organization Role"
                verbose_name_plural = "Organization Roles"

            content_object = models.ForeignKey("myapp.Org", on_delete=models.CASCADE, db_column="org_object_id")
        ```

    See Also:
        - [Django-Guardian Performance Tuning](https://django-guardian.readthedocs.io/en/stable/userguide/performance.html)
        - [How to override the default UserObjectPermission](https://django-guardian.readthedocs.io/en/stable/configuration.html#guardian-user-obj-perms-model)
    """

    group = models.ForeignKey(Group, on_delete=models.CASCADE)

    objects = GroupObjectPermissionManager()

    class Meta:
        abstract = True
        unique_together = ["group", "permission", "content_object"]


class GroupObjectPermissionAbstract(GroupObjectPermissionBase, BaseGenericObjectPermission):
    class Meta(GroupObjectPermissionBase.Meta, BaseGenericObjectPermission.Meta):
        abstract = True
        unique_together = ["group", "permission", "object_pk"]


class GroupObjectPermission(GroupObjectPermissionAbstract):
    """The default implementation of the GroupObjectPermissionAbstract model.

    If `GUARDIAN_GROUP_OBJ_PERMS_MODEL` is not set at the beginning of the project, this model will be used.
    Uses Django's contenttypes framework to store generic relations.

    See Also:
        - [Django's Documentation on Abstract Base Models](https://docs.djangoproject.com/en/stable/topics/db/models/#abstract-base-classes)
        - [Django-Guardian Performance Tuning](https://django-guardian.readthedocs.io/en/stable/userguide/performance.html)
        - [How to override the default GroupObjectPermission](https://django-guardian.readthedocs.io/en/stable/configuration.html#guardian-user-obj-perms-model)
    """

    class Meta(GroupObjectPermissionAbstract.Meta):
        abstract = False
        indexes = [
            models.Index(fields=["permission", "group", "content_type", "object_pk"]),
            models.Index(fields=["group", "content_type", "object_pk"]),
        ]



from django.core.management.base import BaseCommand

from guardian.utils import clean_orphan_obj_perms


class Command(BaseCommand):
    """A wrapper around `guardian.utils.clean_orphan_obj_perms`.

    Seeks and removes all object permissions entries pointing at non-existing targets.
    Returns the number of objects removed.

    Example:
        ```shell
        $ python manage.py clean_orphan_obj_perms
        Removed 11 object permission entries with no targets

        $ python manage.py clean_orphan_obj_perms --batch-size 100 --max-batches 5
        Removed 500 object permission entries with no targets

        $ python manage.py clean_orphan_obj_perms --batch-size 50 --skip-batches 2 --max-duration-secs 60
        Removed 200 object permission entries with no targets
        ```
    """

    help = "Removes object permissions with not existing targets"

    def add_arguments(self, parser):
        parser.add_argument(
            "--batch-size",
            type=int,
            help="Number of objects to process per batch. If not specified, all objects are processed at once.",
        )
        parser.add_argument(
            "--max-batches", type=int, help="Maximum number of batches to process. Use with --batch-size."
        )
        parser.add_argument(
            "--max-duration-secs", type=int, help="Maximum duration in seconds for the cleanup operation."
        )
        parser.add_argument(
            "--skip-batches",
            type=int,
            default=0,
            help="Number of batches to skip before starting cleanup. Use with --batch-size.",
        )

    def handle(self, **options):
        kwargs = {}

        if options["batch_size"] is not None:
            kwargs["batch_size"] = options["batch_size"]

        if options["max_batches"] is not None:
            kwargs["max_batches"] = options["max_batches"]

        if options["max_duration_secs"] is not None:
            kwargs["max_duration_secs"] = options["max_duration_secs"]

        if options["skip_batches"] > 0:
            kwargs["skip_batches"] = options["skip_batches"]

        removed = clean_orphan_obj_perms(**kwargs)

        if options["verbosity"] > 0:
            self.stdout.write("Removed %d object permission entries with no targets" % removed)



from django.conf import settings
from django.core.checks import Tags, Warning, register


# noinspection PyUnusedLocal
@register(Tags.compatibility)
def check_settings(app_configs, **kwargs):
    """Check that settings are implemented properly
    :param app_configs: a list of apps to be checks or None for all
    :param kwargs: keyword arguments
    :return: a list of errors
    """
    checks = []
    if "guardian.backends.ObjectPermissionBackend" not in settings.AUTHENTICATION_BACKENDS:
        msg = (
            "Guardian authentication backend is not hooked. You can add this in settings as eg: "
            "`AUTHENTICATION_BACKENDS = ('django.contrib.auth.backends.ModelBackend', "
            "'guardian.backends.ObjectPermissionBackend')`."
        )
        checks.append(Warning(msg, id="guardian.W001"))
    return checks



from unittest import mock

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase

from guardian.conf import settings as guardian_settings
from guardian.ctypes import get_content_type


class TestConfiguration(TestCase):
    def test_check_configuration(self):
        with mock.patch("guardian.conf.settings.RENDER_403", True):
            with mock.patch("guardian.conf.settings.RAISE_403", True):
                self.assertRaises(ImproperlyConfigured, guardian_settings.check_configuration)

    def test_get_content_type(self):
        with mock.patch(
            "guardian.conf.settings.GET_CONTENT_TYPE", "guardian.testapp.tests.test_conf.get_test_content_type"
        ):
            self.assertEqual(get_content_type(None), "x")


def get_test_content_type(obj):
    """Used in TestConfiguration.test_get_content_type()."""
    return "x"



from unittest import mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import HttpResponse
from django.test import TestCase
from django.test.client import RequestFactory
from django.views.generic import ListView, View

from guardian.mixins import LoginRequiredMixin, PermissionListMixin, PermissionRequiredMixin
from guardian.shortcuts import assign_perm

from ..models import Post


class DatabaseRemovedError(Exception):
    pass


class RemoveDatabaseView(View):
    def get(self, request, *args, **kwargs):
        raise DatabaseRemovedError("You've just allowed db to be removed!")


class PermissionTestView(PermissionRequiredMixin, RemoveDatabaseView):
    permission_required = "testapp.change_post"
    object = None  # should be set at each tests explicitly


class NoObjectView(PermissionRequiredMixin, RemoveDatabaseView):
    permission_required = "testapp.change_post"


class GlobalNoObjectView(PermissionRequiredMixin, RemoveDatabaseView):
    permission_required = "testapp.add_post"
    accept_global_perms = True


class PostPermissionListView(PermissionListMixin, ListView):
    model = Post
    permission_required = "testapp.change_post"
    template_name = "list.html"


class TestViewMixins(TestCase):
    def setUp(self):
        self.post = Post.objects.create(title="foo-post-title")
        self.factory = RequestFactory()
        self.user = get_user_model().objects.create_user("joe", "joe@doe.com", "doe")
        self.client.login(username="joe", password="doe")

    def test_permission_is_checked_before_view_is_computed(self):
        """
        This test would fail if permission is checked **after** view is
        actually resolved.
        """
        request = self.factory.get("/")
        request.user = self.user
        # View.object is set
        view = PermissionTestView.as_view(object=self.post)
        response = view(request)
        self.assertEqual(response.status_code, 302)

        # View.get_object returns object
        PermissionTestView.get_object = lambda instance: self.post
        view = PermissionTestView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 302)
        del PermissionTestView.get_object

    def test_permission_is_checked_before_view_is_computed_perm_denied_raised(self):
        """
        This test would fail if permission is checked **after** view is
        actually resolved.
        """
        request = self.factory.get("/")
        request.user = self.user
        view = PermissionTestView.as_view(raise_exception=True, object=self.post)
        with self.assertRaises(PermissionDenied):
            view(request)

    def test_permission_required_view_configured_wrongly(self):
        """
        This test would fail if permission is checked **after** view is
        actually resolved.
        """
        request = self.factory.get("/")
        request.user = self.user
        request.user.add_obj_perm("change_post", self.post)
        view = PermissionTestView.as_view(permission_required=None, object=self.post)
        with self.assertRaises(ImproperlyConfigured):
            view(request)

    def test_permission_required(self):
        """
        This test would fail if permission is checked **after** view is
        actually resolved.
        """
        request = self.factory.get("/")
        request.user = self.user
        request.user.add_obj_perm("change_post", self.post)
        view = PermissionTestView.as_view(object=self.post)
        with self.assertRaises(DatabaseRemovedError):
            view(request)

    def test_permission_required_no_object(self):
        """
        This test would fail if permission is checked on a view's
        object when it has none
        """

        request = self.factory.get("/")
        request.user = self.user
        request.user.add_obj_perm("change_post", self.post)
        view = NoObjectView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 302)

    def test_permission_required_global_no_object(self):
        """
        This test would fail if permission is checked on a view's
        object when it not set and **no** global permission
        """

        request = self.factory.get("/")
        request.user = self.user
        view = GlobalNoObjectView.as_view()
        response = view(request)
        self.assertEqual(response.status_code, 302)

    def test_permission_granted_global_no_object(self):
        """
        This test would fail if permission is checked on a view's
        object when it not set and **has** global permission
        """

        request = self.factory.get("/")
        request.user = self.user
        assign_perm("testapp.add_post", request.user)
        view = GlobalNoObjectView.as_view()
        with self.assertRaises(DatabaseRemovedError):
            view(request)

    def test_permission_required_as_list(self):
        """
        This test would fail if permission is checked **after** view is
        actually resolved.
        """

        global PermissionTestView

        class SecretView(PermissionTestView):
            on_permission_check_fail = mock.Mock()

        request = self.factory.get("/")
        request.user = self.user
        request.user.add_obj_perm("change_post", self.post)
        SecretView.permission_required = ["testapp.change_post", "testapp.add_post"]
        view = SecretView.as_view(object=self.post)
        response = view(request)
        self.assertEqual(response.status_code, 302)
        SecretView.on_permission_check_fail.assert_called_once_with(request, response, obj=self.post)

        request.user.add_obj_perm("add_post", self.post)
        with self.assertRaises(DatabaseRemovedError):
            view(request)

    def test_login_required_mixin(self):
        class SecretView(LoginRequiredMixin, View):
            redirect_field_name = "foobar"
            login_url = "/let-me-in/"

            def get(self, request):
                return HttpResponse("secret-view")

        request = self.factory.get("/some-secret-page/")
        request.user = AnonymousUser()

        view = SecretView.as_view()

        response = view(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response["Location"], "/let-me-in/?foobar=/some-secret-page/")

        request.user = self.user
        response = view(request)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"secret-view")

    def test_list_permission(self):
        request = self.factory.get("/some-secret-list/")
        request.user = AnonymousUser()

        view = PostPermissionListView.as_view()

        response = view(request)
        self.assertNotContains(response, b"foo-post-title")

        request.user = self.user
        request.user.add_obj_perm("change_post", self.post)

        response = view(request)
        self.assertContains(response, b"foo-post-title")

    def test_any_perm_parameter(self):
        request = self.factory.get("/")
        request.user = self.user
        request.user.add_obj_perm("view_post", self.post)
        self.assertIs(request.user.has_perm("view_post", self.post), True)
        self.assertIs(request.user.has_perm("change_post", self.post), False)
        # success way
        view = PermissionTestView.as_view(
            any_perm=True,
            permission_required=["change_post", "view_post"],
            object=self.post,
        )
        with self.assertRaises(DatabaseRemovedError):
            view(request)
        # fail way
        view = PermissionTestView.as_view(
            any_perm=False,
            permission_required=["change_post", "view_post"],
            object=self.post,
        )
        response = view(request)
        self.assertEqual(response.status_code, 302)



"""
Implementation of per object permissions for Django.
"""

from . import checks  # noqa: F401


def get_version():
    """Return the version string (see pyproject.toml) of the package.

    The value will comply with the [python version specifier format dicted by
    PEP440](https://packaging.python.org/en/latest/specifications/version-specifiers/#version-specifiers)

    Standards for packaging metadata — including version — are defined by PEP 621,
    which specifies how to declare version in pyproject.toml.

    The earlier PEP 396 suggests (but does not mandate) having a __version__
    attribute in __init__.py for the purposes of runtime introspection, but
    it leads to confusion in our development process to define it multiple places.

    PEP 396 has now been revoked, but it is still useful to be able to inspect
    the package version at runtime. This function retains that ability using the
    recommended importlib approach.
    """
    from importlib.metadata import version

    return version("django-guardian")


def monkey_patch_user():
    from django.contrib.auth import get_user_model

    from .utils import evict_obj_perms_cache, get_anonymous_user, get_user_obj_perms_model

    UserObjectPermission = get_user_obj_perms_model()
    User = get_user_model()
    # Prototype User and Group methods
    setattr(User, "get_anonymous", staticmethod(lambda: get_anonymous_user()))
    setattr(User, "add_obj_perm", lambda self, perm, obj: UserObjectPermission.objects.assign_perm(perm, self, obj))
    setattr(User, "del_obj_perm", lambda self, perm, obj: UserObjectPermission.objects.remove_perm(perm, self, obj))
    setattr(User, "evict_obj_perms_cache", evict_obj_perms_cache)


def monkey_patch_group():
    from django.contrib.auth.models import Group

    from .utils import get_group_obj_perms_model

    GroupObjectPermission = get_group_obj_perms_model()
    # Prototype Group methods
    setattr(Group, "add_obj_perm", lambda self, perm, obj: GroupObjectPermission.objects.assign_perm(perm, self, obj))
    setattr(Group, "del_obj_perm", lambda self, perm, obj: GroupObjectPermission.objects.remove_perm(perm, self, obj))



"""
Exceptions used by django-guardian. All internal and guardian-specific errors
should extend GuardianError class.
"""


class GuardianError(Exception):
    """Base class for all guardian-specific exceptions."""

    pass


class NotUserNorGroup(GuardianError):
    """Raised when an object is neither User nor Group."""

    pass


class ObjectNotPersisted(GuardianError):
    """Raised when the object has not been saved to the database."""

    pass


class WrongAppError(GuardianError):
    """Raised when the app name for a permission is incorrect."""

    pass


class MixedContentTypeError(GuardianError):
    """Raised when content type for the provided permissions and/or class do not match."""

    pass


class MultipleIdentityAndObjectError(GuardianError):
    """Raised when an operation is attempted on both user/group and object."""

    pass



from itertools import chain

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.management import create_permissions
from django.contrib.auth.models import AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from guardian.conf import settings as guardian_settings
from guardian.core import ObjectPermissionChecker
from guardian.exceptions import NotUserNorGroup
from guardian.management import create_anonymous_user
from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.shortcuts import assign_perm
from guardian.testapp.models import Project, ProjectGroupObjectPermission, ProjectUserObjectPermission
from guardian.utils import evict_obj_perms_cache

auth_app = django_apps.get_app_config("auth")
User = get_user_model()


class CustomUserTests(TestCase):
    def test_create_anonymous_user(self):
        create_anonymous_user(object(), using="default")
        self.assertEqual(1, User.objects.all().count())
        anonymous = User.objects.all()[0]
        self.assertEqual(anonymous.username, guardian_settings.ANONYMOUS_USER_NAME)


class ObjectPermissionTestCase(TestCase):
    def setUp(self):
        self.group, created = Group.objects.get_or_create(name="jackGroup")
        self.user, created = User.objects.get_or_create(username="jack")
        self.user.groups.add(self.group)
        self.ctype = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")
        self.ctype_qset = ContentType.objects.filter(model="bar", app_label="fake-for-guardian-tests")
        self.ctype_list = [self.ctype_qset.first()]
        self.anonymous_user = User.objects.get(username=guardian_settings.ANONYMOUS_USER_NAME)

    def get_permission(self, codename, app_label=None):
        qs = Permission.objects
        if app_label:
            qs = qs.filter(content_type__app_label=app_label)
        return Permission.objects.get(codename=codename)


class ObjectPermissionCheckerTest(ObjectPermissionTestCase):
    def setUp(self):
        super().setUp()
        # Required if MySQL backend is used :/
        create_permissions(auth_app, 1)

    def test_cache_for_queries_count(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            checker = ObjectPermissionChecker(self.user)

            # has_perm on Checker should spawn only two queries plus one extra
            # for fetching the content type first time we check for specific
            # model and two more content types as there are additional checks
            # at get_user_obj_perms_model and get_group_obj_perms_model
            query_count = len(connection.queries)
            res = checker.has_perm("change_group", self.group)
            if "guardian.testapp" in settings.INSTALLED_APPS:
                expected = 5
            else:
                # TODO: This is strange, need to investigate; totally not sure
                # why there are more queries if testapp is not included
                expected = 11
            self.assertEqual(len(connection.queries), query_count + expected)

            # Checking again shouldn't spawn any queries
            query_count = len(connection.queries)
            res_new = checker.has_perm("change_group", self.group)
            self.assertEqual(res, res_new)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            query_count = len(connection.queries)
            checker.has_perm("delete_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance should spawn 2 queries
            new_group = Group.objects.create(name="new-group")
            query_count = len(connection.queries)
            checker.has_perm("change_group", new_group)
            self.assertEqual(len(connection.queries), query_count + 2)

            # Checking for permission for other model should spawn 4 queries
            # every added direct relation adds one more query..
            # (again: content type and actual permissions for the object...
            query_count = len(connection.queries)
            checker.has_perm("change_user", self.user)
            self.assertEqual(len(connection.queries), query_count + 4)

        finally:
            settings.DEBUG = False

    def test_init(self):
        self.assertRaises(NotUserNorGroup, ObjectPermissionChecker, user_or_group=ContentType())
        self.assertRaises(NotUserNorGroup, ObjectPermissionChecker)

    def test_anonymous_user(self):
        user = AnonymousUser()
        check = ObjectPermissionChecker(user)
        # assert anonymous user has no object permissions at all for obj
        self.assertTrue([] == list(check.get_perms(self.ctype)))

    def test_superuser(self):
        user = User.objects.create(username="superuser", is_superuser=True)
        check = ObjectPermissionChecker(user)
        ctype = ContentType.objects.get_for_model(self.ctype)
        perms = sorted(chain(*Permission.objects.filter(content_type=ctype).values_list("codename")))
        self.assertEqual(perms, check.get_perms(self.ctype))
        for perm in perms:
            self.assertTrue(check.has_perm(perm, self.ctype))

    def test_not_active_superuser(self):
        user = User.objects.create(username="not_active_superuser", is_superuser=True, is_active=False)
        check = ObjectPermissionChecker(user)
        ctype = ContentType.objects.get_for_model(self.ctype)
        perms = sorted(chain(*Permission.objects.filter(content_type=ctype).values_list("codename")))
        self.assertEqual(check.get_perms(self.ctype), [])
        for perm in perms:
            self.assertFalse(check.has_perm(perm, self.ctype))

    def test_not_active_user(self):
        user = User.objects.create(username="notactive")
        assign_perm("change_contenttype", user, self.ctype)

        # new ObjectPermissionChecker is created for each User.has_perm call
        self.assertTrue(user.has_perm("change_contenttype", self.ctype))
        user.is_active = False
        self.assertFalse(user.has_perm("change_contenttype", self.ctype))

        # use on one checker only (as user's is_active attr should be checked
        # before try to use cache
        user = User.objects.create(username="notactive-cache")
        assign_perm("change_contenttype", user, self.ctype)

        check = ObjectPermissionChecker(user)
        self.assertTrue(check.has_perm("change_contenttype", self.ctype))
        user.is_active = False
        self.assertFalse(check.has_perm("change_contenttype", self.ctype))

    def test_get_perms(self):
        group = Group.objects.create(name="group")
        obj1 = ContentType.objects.create(model="foo", app_label="guardian-tests")
        obj2 = ContentType.objects.create(model="bar", app_label="guardian-tests")

        assign_perms = {
            group: ("change_group", "delete_group"),
            obj1: ("change_contenttype", "delete_contenttype"),
            obj2: ("delete_contenttype",),
        }

        check = ObjectPermissionChecker(self.user)

        for obj, perms in assign_perms.items():
            for perm in perms:
                UserObjectPermission.objects.assign_perm(perm, self.user, obj)
            self.assertEqual(sorted(perms), sorted(check.get_perms(obj)))

        check = ObjectPermissionChecker(self.group)

        for obj, perms in assign_perms.items():
            for perm in perms:
                GroupObjectPermission.objects.assign_perm(perm, self.group, obj)
            self.assertEqual(sorted(perms), sorted(check.get_perms(obj)))

    def test_prefetch_user_perms(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group1 = Group.objects.create(name="group1")
            group2 = Group.objects.create(name="group2")
            user = User.objects.create(username="active_user", is_active=True)
            assign_perm("change_group", user, self.group)
            assign_perm("change_group", user, group1)
            checker = ObjectPermissionChecker(user)

            # Prefetch permissions
            prefetched_objects = [self.group, group1, group2]
            self.assertTrue(checker.prefetch_perms(prefetched_objects))
            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(prefetched_objects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            checker.has_perm("change_group", group1)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on Group2, we still should
            #  not hit DB
            self.assertFalse(checker.has_perm("change_group", group2))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_prefetch_superuser_perms(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group1 = Group.objects.create(name="group1")
            user = User.objects.create(username="active_superuser", is_superuser=True, is_active=True)
            assign_perm("change_group", user, self.group)
            checker = ObjectPermissionChecker(user)

            # Prefetch permissions
            prefetched_objects = [self.group, group1]
            self.assertTrue(checker.prefetch_perms(prefetched_objects))
            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(prefetched_objects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            checker.has_perm("change_group", group1)
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_prefetch_group_perms(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group1 = Group.objects.create(name="group1")
            group2 = Group.objects.create(name="group2")
            assign_perm("change_group", group1, self.group)
            assign_perm("change_group", group1, group1)
            checker = ObjectPermissionChecker(group1)

            # Prefetch permissions
            prefetched_objects = [self.group, group1, group2]
            self.assertTrue(checker.prefetch_perms(prefetched_objects))

            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(prefetched_objects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_group", self.group)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            checker.has_perm("change_group", group1)
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on Group2, we still should
            #  not hit DB
            self.assertFalse(checker.has_perm("change_group", group2))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_prefetch_user_perms_direct_rel(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            user = User.objects.create(username="active_user", is_active=True)
            projects = [Project.objects.create(name="Project%s" % i) for i in range(3)]
            assign_perm("change_project", user, projects[0])
            assign_perm("change_project", user, projects[1])

            checker = ObjectPermissionChecker(user)

            # Prefetch permissions
            self.assertTrue(checker.prefetch_perms(projects))
            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(projects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any
            #  queries
            checker.has_perm("change_project", projects[1])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on projects[2], we still
            #  should not hit DB
            self.assertFalse(checker.has_perm("change_project", projects[2]))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_prefetch_superuser_perms_direct_rel(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            user = User.objects.create(username="active_user", is_active=True, is_superuser=True)
            projects = [Project.objects.create(name="Project%s" % i) for i in range(2)]
            assign_perm("change_project", user, projects[0])

            checker = ObjectPermissionChecker(user)

            # Prefetch permissions
            self.assertTrue(checker.prefetch_perms(projects))
            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(projects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any
            #  queries
            checker.has_perm("change_project", projects[1])
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_prefetch_group_perms_direct_rel(self):
        settings.DEBUG = True
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group = Group.objects.create(name="new-group")
            projects = [Project.objects.create(name="Project%s" % i) for i in range(3)]
            assign_perm("change_project", group, projects[0])
            assign_perm("change_project", group, projects[1])

            checker = ObjectPermissionChecker(group)

            # Prefetch permissions
            self.assertTrue(checker.prefetch_perms(projects))
            query_count = len(connection.queries)

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(projects))

            # Checking shouldn't spawn any queries
            checker.has_perm("change_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            checker.has_perm("delete_project", projects[0])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any
            #  queries
            checker.has_perm("change_project", projects[1])
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on projects[2], we still
            #  should not hit DB
            self.assertFalse(checker.has_perm("change_project", projects[2]))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False

    def test_autoprefetch_user_perms(self):
        settings.DEBUG = True
        guardian_settings.AUTO_PREFETCH = True
        ProjectUserObjectPermission.enabled = False
        ProjectGroupObjectPermission.enabled = False
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group1 = Group.objects.create(name="group1")
            group2 = Group.objects.create(name="group2")
            user = User.objects.create(username="active_user", is_active=True)
            assign_groups = [self.group, group1]
            for group in assign_groups:
                assign_perm("change_group", user, group)

            # Faux monkeypatching
            setattr(User, "evict_obj_perms_cache", evict_obj_perms_cache)

            # Try to evict non-existent cache, should return false
            self.assertFalse(user.evict_obj_perms_cache())

            checker = ObjectPermissionChecker(user)

            pre_query_count = len(connection.queries)
            # Fill cache
            checker._prefetch_cache()
            query_count = len(connection.queries)

            # Ensure two queries have been made
            self.assertEqual(query_count - pre_query_count, 2)

            # Check user has cache attribute
            self.assertTrue(hasattr(user, "_guardian_perms_cache"))

            # Checking cache is filled
            self.assertEqual(len(checker._obj_perms_cache), len(assign_groups))

            # Checking shouldn't spawn any queries
            self.assertTrue(checker.has_perm("change_group", self.group))
            self.assertTrue(user.has_perm("change_group", self.group))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            self.assertFalse(checker.has_perm("delete_group", self.group))
            self.assertFalse(user.has_perm("delete_group", self.group))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            self.assertTrue(checker.has_perm("change_group", group1))
            self.assertTrue(user.has_perm("change_group", group1))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on Group2, we still should
            #  not hit DB
            self.assertFalse(checker.has_perm("change_group", group2))
            self.assertFalse(user.has_perm("change_group", group2))
            self.assertEqual(len(connection.queries), query_count)

            # Evict cache and verify that reloaded perms have changed
            self.assertTrue(user.evict_obj_perms_cache())
            self.assertFalse(hasattr(user, "_guardian_perms_cache"))
            assign_perm("delete_group", user, self.group)
            query_count = len(connection.queries)
            # New checker object so that we reload perms from db
            checker = ObjectPermissionChecker(user)
            self.assertTrue(checker.has_perm("delete_group", self.group))
            self.assertTrue(user.has_perm("delete_group", self.group))
            # Two more queries should have been executed
            self.assertEqual(len(connection.queries), query_count + 2)

        finally:
            settings.DEBUG = False
            guardian_settings.AUTO_PREFETCH = False
            ProjectUserObjectPermission.enabled = True
            ProjectGroupObjectPermission.enabled = True

    def test_autoprefetch_superuser_perms(self):
        settings.DEBUG = True
        guardian_settings.AUTO_PREFETCH = True
        ProjectUserObjectPermission.enabled = False
        ProjectGroupObjectPermission.enabled = False
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group1 = Group.objects.create(name="group1")
            user = User.objects.create(username="active_superuser", is_superuser=True, is_active=True)
            assign_perm("change_group", user, self.group)
            checker = ObjectPermissionChecker(user)

            # Fill cache
            checker._prefetch_cache()
            query_count = len(connection.queries)

            # Check user has cache attribute
            self.assertTrue(hasattr(user, "_guardian_perms_cache"))

            # Checking shouldn't spawn any queries
            self.assertTrue(checker.has_perm("change_group", self.group))
            self.assertTrue(user.has_perm("change_group", self.group))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            self.assertTrue(checker.has_perm("delete_group", self.group))
            self.assertTrue(user.has_perm("delete_group", self.group))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            self.assertTrue(checker.has_perm("change_group", group1))
            self.assertTrue(user.has_perm("change_group", group1))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False
            guardian_settings.AUTO_PREFETCH = False
            ProjectUserObjectPermission.enabled = True
            ProjectGroupObjectPermission.enabled = True

    def test_autoprefetch_group_perms(self):
        settings.DEBUG = True
        guardian_settings.AUTO_PREFETCH = True
        ProjectUserObjectPermission.enabled = False
        ProjectGroupObjectPermission.enabled = False
        try:
            from django.db import connection

            ContentType.objects.clear_cache()
            group = Group.objects.create(name="new-group")
            projects = [Project.objects.create(name="Project%s" % i) for i in range(3)]
            assign_perm("change_project", group, projects[0])
            assign_perm("change_project", group, projects[1])

            checker = ObjectPermissionChecker(group)

            pre_query_count = len(connection.queries)
            # Fill cache
            checker._prefetch_cache()
            query_count = len(connection.queries)

            # Ensure only one query has been made
            self.assertEqual(query_count - pre_query_count, 1)

            # Check group has cache attribute
            self.assertTrue(hasattr(group, "_guardian_perms_cache"))

            # Checking shouldn't spawn any queries
            self.assertTrue(checker.has_perm("change_project", projects[0]))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for other permission but for Group object again
            # shouldn't spawn any query too
            self.assertFalse(checker.has_perm("delete_project", projects[0]))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any
            #  queries
            self.assertTrue(checker.has_perm("change_project", projects[1]))
            self.assertEqual(len(connection.queries), query_count)

            # Checking for same model but other instance shouldn't spawn any queries
            # Even though User doesn't have perms on projects[2], we still
            #  should not hit DB
            self.assertFalse(checker.has_perm("change_project", projects[2]))
            self.assertEqual(len(connection.queries), query_count)
        finally:
            settings.DEBUG = False
            guardian_settings.AUTO_PREFETCH = False
            ProjectUserObjectPermission.enabled = True
            ProjectGroupObjectPermission.enabled = True



from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.test import TestCase

from guardian.exceptions import NotUserNorGroup
from guardian.models import GroupObjectPermission, UserObjectPermission, UserObjectPermissionBase
from guardian.testapp.models import Project, ProjectGroupObjectPermission, ProjectUserObjectPermission
from guardian.testapp.tests.conf import skipUnlessTestApp
from guardian.testapp.tests.test_core import ObjectPermissionTestCase
from guardian.utils import (
    get_anonymous_user,
    get_group_obj_perms_model,
    get_identity,
    get_obj_perms_model,
    get_user_obj_perms_model,
)

User = get_user_model()


class GetAnonymousUserTest(TestCase):
    def test(self):
        anon = get_anonymous_user()
        self.assertTrue(isinstance(anon, User))


class GetIdentityTest(ObjectPermissionTestCase):
    def test_user(self):
        user, group = get_identity(self.user)
        self.assertTrue(isinstance(user, User))
        self.assertEqual(group, None)

    def test_anonymous_user(self):
        anon = AnonymousUser()
        user, group = get_identity(anon)
        self.assertTrue(isinstance(user, User))
        self.assertEqual(group, None)

    def test_group(self):
        user, group = get_identity(self.group)
        self.assertTrue(isinstance(group, Group))
        self.assertEqual(user, None)

    def test_not_user_nor_group(self):
        self.assertRaises(NotUserNorGroup, get_identity, 1)
        self.assertRaises(NotUserNorGroup, get_identity, "User")
        self.assertRaises(NotUserNorGroup, get_identity, User)

    def test_multiple_user_qs(self):
        user, group = get_identity(User.objects.all())
        self.assertIsInstance(user, models.QuerySet)
        self.assertIsNone(group)

    def test_multiple_user_list(self):
        user, group = get_identity([self.user])
        self.assertIsInstance(user, list)
        self.assertIsNone(group)

    def test_multiple_group_qs(self):
        user, group = get_identity(Group.objects.all())
        self.assertIsInstance(group, models.QuerySet)
        self.assertIsNone(user)

    def test_multiple_group_list(self):
        user, group = get_identity([self.group])
        self.assertIsInstance(group, list)
        self.assertIsNone(user)


@skipUnlessTestApp
class GetUserObjPermsModelTest(TestCase):
    def test_for_instance(self):
        project = Project(name="Foobar")
        self.assertEqual(get_user_obj_perms_model(project), ProjectUserObjectPermission)

    def test_for_class(self):
        self.assertEqual(get_user_obj_perms_model(Project), ProjectUserObjectPermission)

    def test_default(self):
        self.assertEqual(get_user_obj_perms_model(ContentType), UserObjectPermission)

    def test_user_model(self):
        # this test assumes that there were no direct obj perms model to User
        # model defined (i.e. while testing guardian app in some custom
        # project)
        self.assertEqual(get_user_obj_perms_model(User), UserObjectPermission)


@skipUnlessTestApp
class GetGroupObjPermsModelTest(TestCase):
    def test_for_instance(self):
        project = Project(name="Foobar")
        self.assertEqual(get_group_obj_perms_model(project), ProjectGroupObjectPermission)

    def test_for_class(self):
        self.assertEqual(get_group_obj_perms_model(Project), ProjectGroupObjectPermission)

    def test_default(self):
        self.assertEqual(get_group_obj_perms_model(ContentType), GroupObjectPermission)

    def test_group_model(self):
        # this test assumes that there were no direct obj perms model to Group
        # model defined (i.e. while testing guardian app in some custom
        # project)
        self.assertEqual(get_group_obj_perms_model(Group), GroupObjectPermission)


class GetObjPermsModelTest(TestCase):
    def test_image_field(self):
        class SomeModel(models.Model):
            image = models.FileField(upload_to="images/")

        obj = SomeModel()
        perm_model = get_obj_perms_model(obj, UserObjectPermissionBase, UserObjectPermission)
        self.assertEqual(perm_model, UserObjectPermission)

    def test_file_field(self):
        class SomeModel2(models.Model):
            file = models.FileField(upload_to="images/")

        obj = SomeModel2()
        perm_model = get_obj_perms_model(obj, UserObjectPermissionBase, UserObjectPermission)
        self.assertEqual(perm_model, UserObjectPermission)



# Generated by Django 5.2.5 on 2025-08-28 23:39

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        ("testapp", "0007_genericgroupobjectpermission"),
    ]

    operations = [
        migrations.AlterField(
            model_name="project",
            name="created_at",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from unittest import mock

from django.conf import global_settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser, Group
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db.models.base import ModelBase
from django.http import HttpRequest, HttpResponse, HttpResponseForbidden, HttpResponseNotFound, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import TemplateDoesNotExist
from django.test import TestCase

from guardian.compat import get_user_model_path, get_user_permission_full_codename
from guardian.decorators import permission_required, permission_required_or_403, permission_required_or_404
from guardian.exceptions import GuardianError, WrongAppError
from guardian.shortcuts import assign_perm
from guardian.testapp.tests.conf import TestDataMixin, override_settings, skipUnlessTestApp

User = get_user_model()
user_model_path = get_user_model_path()


@skipUnlessTestApp
class PermissionRequiredTest(TestDataMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.anon = AnonymousUser()
        self.user = User.objects.get_or_create(username="jack")[0]
        self.group = Group.objects.get_or_create(name="jackGroup")[0]

    def _get_request(self, user=None):
        if user is None:
            user = AnonymousUser()
        request = HttpRequest()
        request.user = user
        return request

    def test_no_args(self):
        try:

            @permission_required
            def dummy_view(request):
                return HttpResponse("dummy_view")
        except GuardianError:
            pass
        else:
            self.fail(
                "Trying to decorate using permission_required without "
                "permission as first argument should raise exception"
            )

    def test_RENDER_403_is_false(self):
        request = self._get_request(self.anon)

        @permission_required_or_403("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.RENDER_403", False):
            response = dummy_view(request)
            self.assertEqual(response.content, b"")
            self.assertTrue(isinstance(response, HttpResponseForbidden))

    def test_RENDER_404_is_false(self):
        request = self._get_request(self.anon)

        @permission_required_or_404("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.RENDER_404", False):
            response = dummy_view(request)
            self.assertEqual(response.content, b"")
            self.assertTrue(isinstance(response, HttpResponseNotFound))

    @mock.patch("guardian.conf.settings.RENDER_403", True)
    def test_TEMPLATE_403_setting(self):
        request = self._get_request(self.anon)

        @permission_required_or_403("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.TEMPLATE_403", "dummy403.html"):
            response = dummy_view(request)
            self.assertEqual(response.content, b"foobar403\n")

    @mock.patch("guardian.conf.settings.RENDER_404", True)
    def test_TEMPLATE_404_setting(self):
        request = self._get_request(self.anon)

        @permission_required_or_404("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.TEMPLATE_404", "dummy404.html"):
            response = dummy_view(request)
            self.assertEqual(response.content, b"foobar404\n")

    @mock.patch("guardian.conf.settings.RENDER_403", True)
    def test_403_response_raises_error(self):
        request = self._get_request(self.anon)

        @permission_required_or_403("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.TEMPLATE_403", "_non-exisitng-403.html"):
            self.assertRaises(TemplateDoesNotExist, dummy_view, request)

    @mock.patch("guardian.conf.settings.RENDER_404", True)
    def test_404_response_raises_error(self):
        request = self._get_request(self.anon)

        @permission_required_or_404("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        with mock.patch("guardian.conf.settings.TEMPLATE_404", "_non-exisitng-404.html"):
            self.assertRaises(TemplateDoesNotExist, dummy_view, request)

    @mock.patch("guardian.conf.settings.RENDER_403", False)
    @mock.patch("guardian.conf.settings.RAISE_403", True)
    def test_RAISE_403_setting_is_true(self):
        request = self._get_request(self.anon)

        @permission_required_or_403("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertRaises(PermissionDenied, dummy_view, request)

    @mock.patch("guardian.conf.settings.RENDER_404", False)
    @mock.patch("guardian.conf.settings.RAISE_404", True)
    def test_RAISE_404_setting_is_true(self):
        request = self._get_request(self.anon)

        @permission_required_or_404("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertRaises(ObjectDoesNotExist, dummy_view, request)

    def test_anonymous_user_wrong_app(self):
        request = self._get_request(self.anon)

        @permission_required_or_403("not_installed_app.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertEqual(dummy_view(request).status_code, 403)

    def test_anonymous_user_wrong_codename(self):
        request = self._get_request()

        @permission_required_or_403("auth.wrong_codename")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertEqual(dummy_view(request).status_code, 403)

    def test_anonymous_user(self):
        request = self._get_request()

        @permission_required_or_403("auth.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertEqual(dummy_view(request).status_code, 403)

    def test_wrong_lookup_variables_number(self):
        request = self._get_request()

        try:

            @permission_required_or_403("auth.change_user", (User, "username"))
            def dummy_view(request, username):
                pass

            dummy_view(request, username="jack")
        except GuardianError:
            pass
        else:
            self.fail(
                "If lookup variables are passed they must be tuple of: "
                "(ModelClass/app_label.ModelClass/queryset, "
                "<pair of lookup_string and view_arg>)\n"
                "Otherwise GuardianError should be raised"
            )

    def test_wrong_lookup_variables(self):
        request = self._get_request()

        args = (
            (2010, "username", "username"),
            ("User", "username", "username"),
            (User, "username", "no_arg"),
        )
        for tup in args:
            try:

                @permission_required_or_403("auth.change_user", tup)
                def show_user(request, username):
                    user = get_object_or_404(User, username=username)
                    return HttpResponse("It's %s here!" % user.username)

                show_user(request, "jack")
            except GuardianError:
                pass
            else:
                self.fail("Wrong arguments given but GuardianError not raised")

    def test_user_has_no_access(self):
        request = self._get_request()

        @permission_required_or_403("auth.change_user")
        def dummy_view(request):
            return HttpResponse("dummy_view")

        self.assertEqual(dummy_view(request).status_code, 403)

    def test_user_has_access(self):
        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")
        assign_perm(perm, self.user, obj=joe)

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (user_model_path, "username", "username"))
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"dummy_view")

    def test_user_has_access_on_model_with_metaclass(self):
        """
        Test to the fix issues of comparison made via type()
        in the decorator. In the case of a `Model` implementing
        a custom metaclass, the decorator fail because type
        doesn't return `ModelBase`
        """
        perm = get_user_permission_full_codename("change")

        class TestMeta(ModelBase):
            pass

        class ProxyUser(User):
            class Meta:
                proxy = True
                app_label = User._meta.app_label

            __metaclass__ = TestMeta

        joe, created = ProxyUser.objects.get_or_create(username="joe")
        assign_perm(perm, self.user, obj=joe)

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (ProxyUser, "username", "username"))
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"dummy_view")

    def test_user_has_obj_access_even_if_we_also_check_for_global(self):
        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")
        assign_perm(perm, self.user, obj=joe)

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (user_model_path, "username", "username"), accept_global_perms=True)
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"dummy_view")

    def test_user_has_no_obj_perm_access(self):
        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (user_model_path, "username", "username"))
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 403)

    def test_user_has_global_perm_access_but_flag_not_set(self):
        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")
        assign_perm(perm, self.user)

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (user_model_path, "username", "username"))
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 403)

    def test_user_has_global_perm_access(self):
        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")
        assign_perm(perm, self.user)

        request = self._get_request(self.user)

        @permission_required_or_403(perm, (user_model_path, "username", "username"), accept_global_perms=True)
        def dummy_view(request, username):
            return HttpResponse("dummy_view")

        response = dummy_view(request, username="joe")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"dummy_view")

    def test_model_lookup(self):
        request = self._get_request(self.user)

        perm = get_user_permission_full_codename("change")
        joe, created = User.objects.get_or_create(username="joe")
        assign_perm(perm, self.user, obj=joe)

        models = (
            user_model_path,
            User,
            User.objects.filter(is_active=True),
        )
        for model in models:

            @permission_required_or_403(perm, (model, "username", "username"))
            def dummy_view(request, username):
                get_object_or_404(User, username=username)
                return HttpResponse("hello")

            response = dummy_view(request, username=joe.username)
            self.assertEqual(response.content, b"hello")

    def test_redirection_raises_wrong_app_error(self):
        from guardian.testapp.models import Project

        request = self._get_request(self.user)

        User.objects.create(username="foo")
        Project.objects.create(name="foobar")

        @permission_required("auth.change_group", (Project, "name", "group_name"), login_url="/foobar/")
        def dummy_view(request, project_name):
            pass

        # 'auth.change_group' is wrong permission codename (should be one
        # related with User
        self.assertRaises(WrongAppError, dummy_view, request, group_name="foobar")

    def test_redirection(self):
        from guardian.testapp.models import Project

        request = self._get_request(self.user)

        User.objects.create(username="foo")
        Project.objects.create(name="foobar")

        @permission_required("testapp.change_project", (Project, "name", "project_name"), login_url="/foobar/")
        def dummy_view(request, project_name):
            pass

        response = dummy_view(request, project_name="foobar")
        self.assertTrue(isinstance(response, HttpResponseRedirect))
        self.assertTrue(response.headers["location"].startswith("/foobar/"))

    @override_settings(LOGIN_URL="django.contrib.auth.views.login")
    def test_redirection_class(self):
        view_url = "/permission_required/"

        response = self.client.get(view_url)
        # this should be '/account/login'
        self.assertRedirects(response, global_settings.LOGIN_URL + "?next=" + view_url)



from django.test import TestCase

from guardian.checks import check_settings


class SystemCheckTestCase(TestCase):
    def test_checks(self):
        """Test custom system checks
        :return: None
        """
        self.assertFalse(check_settings(None))

        with self.settings(
            AUTHENTICATION_BACKENDS=("django.contrib.auth.backends.ModelBackend",),
        ):
            self.assertEqual(len(check_settings(None)), 1)



from typing import Any, Type, Union

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.utils.module_loading import import_string

from guardian.conf import settings as guardian_settings


def get_content_type(obj: Union[Model, Type[Model]]) -> Any:
    get_content_type_function = import_string(guardian_settings.GET_CONTENT_TYPE)
    return get_content_type_function(obj)


def get_default_content_type(obj: Union[Model, Type[Model]]) -> ContentType:
    """Get content type for a given object using Django's content type framework.

    Parameters:
        obj (Model | Type): Object for which content type is to be fetched.

    Returns:
        Content type for the given object.

    See Also:
        https://docs.djangoproject.com/en/5.1/ref/contrib/contenttypes/

    """
    return ContentType.objects.get_for_model(obj)



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.test import TestCase

from guardian.shortcuts import (
    assign_perm,
    get_groups_with_perms,
    get_objects_for_group,
    get_objects_for_user,
    get_users_with_perms,
    remove_perm,
)
from guardian.testapp.models import (
    Mixed,
    Project,
    ProjectGroupObjectPermission,
    ProjectUserObjectPermission,
    ReverseMixed,
)
from guardian.testapp.tests.conf import skipUnlessTestApp

User = get_user_model()


@skipUnlessTestApp
class TestDirectUserPermissions(TestCase):
    def setUp(self):
        self.joe = User.objects.create_user("joe", "joe@example.com", "foobar")
        self.project = Project.objects.create(name="Foobar")

    def get_perm(self, codename):
        filters = {"content_type__app_label": "testapp", "codename": codename}
        return Permission.objects.get(**filters)

    def test_after_perm_is_created_without_shortcut(self):
        perm = self.get_perm("add_project")
        # we should not use assign here - if generic user obj perms model is
        # used then everything could go fine if using assign shortcut and we
        # would not be able to see any problem
        ProjectUserObjectPermission.objects.create(
            user=self.joe,
            permission=perm,
            content_object=self.project,
        )
        self.assertTrue(self.joe.has_perm("add_project", self.project))

    def test_assign_perm(self):
        assign_perm("add_project", self.joe, self.project)
        filters = {
            "content_object": self.project,
            "permission__codename": "add_project",
            "user": self.joe,
        }
        result = ProjectUserObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 1)

    def test_remove_perm(self):
        assign_perm("add_project", self.joe, self.project)
        filters = {
            "content_object": self.project,
            "permission__codename": "add_project",
            "user": self.joe,
        }
        result = ProjectUserObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 1)

        remove_perm("add_project", self.joe, self.project)
        result = ProjectUserObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 0)

    def test_get_users_with_perms(self):
        User.objects.create_user("john", "john@foobar.com", "john")
        jane = User.objects.create_user("jane", "jane@foobar.com", "jane")
        assign_perm("add_project", self.joe, self.project)
        assign_perm("change_project", self.joe, self.project)
        assign_perm("change_project", jane, self.project)
        self.assertEqual(
            get_users_with_perms(self.project, attach_perms=True),
            {
                self.joe: ["add_project", "change_project"],
                jane: ["change_project"],
            },
        )

    def test_get_users_with_perms_plus_groups(self):
        User.objects.create_user("john", "john@foobar.com", "john")
        jane = User.objects.create_user("jane", "jane@foobar.com", "jane")
        group = Group.objects.create(name="devs")
        self.joe.groups.add(group)
        assign_perm("add_project", self.joe, self.project)
        assign_perm("change_project", group, self.project)
        assign_perm("change_project", jane, self.project)
        self.assertEqual(
            get_users_with_perms(self.project, attach_perms=True),
            {
                self.joe: ["add_project", "change_project"],
                jane: ["change_project"],
            },
        )

    def test_get_objects_for_user(self):
        foo = Project.objects.create(name="foo")
        bar = Project.objects.create(name="bar")
        assign_perm("add_project", self.joe, foo)
        assign_perm("add_project", self.joe, bar)
        assign_perm("change_project", self.joe, bar)

        result = get_objects_for_user(self.joe, "testapp.add_project")
        self.assertEqual(sorted(p.pk for p in result), sorted([foo.pk, bar.pk]))

    def test_get_all_permissions(self):
        foo = Project.objects.create(name="foo")
        assign_perm("add_project", self.joe, foo)
        assign_perm("change_project", self.joe, foo)

        result = self.joe.get_all_permissions(foo)
        self.assertEqual(result, {"add_project", "change_project"})

    def test_get_all_permissions_no_object(self):
        foo = Project.objects.create(name="foo")
        assign_perm("add_project", self.joe, foo)
        assign_perm("change_project", self.joe, foo)

        result = self.joe.get_all_permissions()
        self.assertEqual(result, set())


@skipUnlessTestApp
class TestDirectGroupPermissions(TestCase):
    def setUp(self):
        self.joe = User.objects.create_user("joe", "joe@example.com", "foobar")
        self.group = Group.objects.create(name="admins")
        self.joe.groups.add(self.group)
        self.project = Project.objects.create(name="Foobar")

    def get_perm(self, codename):
        filters = {"content_type__app_label": "testapp", "codename": codename}
        return Permission.objects.get(**filters)

    def test_after_perm_is_created_without_shortcut(self):
        perm = self.get_perm("add_project")
        # we should not use assign here - if generic user obj perms model is
        # used then everything could go fine if using assign shortcut and we
        # would not be able to see any problem
        ProjectGroupObjectPermission.objects.create(
            group=self.group,
            permission=perm,
            content_object=self.project,
        )
        self.assertTrue(self.joe.has_perm("add_project", self.project))

    def test_assign_perm(self):
        assign_perm("add_project", self.group, self.project)
        filters = {
            "content_object": self.project,
            "permission__codename": "add_project",
            "group": self.group,
        }
        result = ProjectGroupObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 1)

    def test_remove_perm(self):
        assign_perm("add_project", self.group, self.project)
        filters = {
            "content_object": self.project,
            "permission__codename": "add_project",
            "group": self.group,
        }
        result = ProjectGroupObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 1)

        remove_perm("add_project", self.group, self.project)
        result = ProjectGroupObjectPermission.objects.filter(**filters).count()
        self.assertEqual(result, 0)

    def test_get_groups_with_perms(self):
        Group.objects.create(name="managers")
        devs = Group.objects.create(name="devs")
        assign_perm("add_project", self.group, self.project)
        assign_perm("change_project", self.group, self.project)
        assign_perm("change_project", devs, self.project)
        self.assertEqual(
            get_groups_with_perms(self.project, attach_perms=True),
            {
                self.group: ["add_project", "change_project"],
                devs: ["change_project"],
            },
        )

    def test_get_groups_with_perms_doesnt_spawn_extra_queries_for_more_groups_with_perms(self):
        Group.objects.create(name="managers")
        devs = Group.objects.create(name="devs")
        devs1 = Group.objects.create(name="devs1")
        devs2 = Group.objects.create(name="devs2")
        devs3 = Group.objects.create(name="devs3")
        devs4 = Group.objects.create(name="devs4")
        devs5 = Group.objects.create(name="devs5")
        assign_perm("add_project", self.group, self.project)
        assign_perm("change_project", self.group, self.project)
        for group in [devs, devs1, devs2, devs3, devs4, devs5]:
            assign_perm("add_project", group, self.project)
            assign_perm("change_project", group, self.project)

        with self.assertNumQueries(3):
            result = get_groups_with_perms(self.project, attach_perms=True)

        self.assertEqual(
            result,
            {
                self.group: ["add_project", "change_project"],
                devs: ["add_project", "change_project"],
                devs1: ["add_project", "change_project"],
                devs2: ["add_project", "change_project"],
                devs3: ["add_project", "change_project"],
                devs4: ["add_project", "change_project"],
                devs5: ["add_project", "change_project"],
            },
        )

    def test_get_objects_for_group(self):
        foo = Project.objects.create(name="foo")
        bar = Project.objects.create(name="bar")
        assign_perm("add_project", self.group, foo)
        assign_perm("add_project", self.group, bar)
        assign_perm("change_project", self.group, bar)

        result = get_objects_for_group(self.group, "testapp.add_project")
        self.assertEqual(sorted(p.pk for p in result), sorted([foo.pk, bar.pk]))


@skipUnlessTestApp
class TestMixedDirectAndGenericObjectPermission(TestCase):
    def setUp(self):
        self.joe = User.objects.create_user("joe", "joe@example.com", "foobar")
        self.group = Group.objects.create(name="admins")
        self.joe.groups.add(self.group)
        self.mixed = Mixed.objects.create(name="Foobar")
        self.reverse_mixed = ReverseMixed.objects.create(name="Foobar")

    def test_get_users_with_perms_plus_groups(self):
        User.objects.create_user("john", "john@foobar.com", "john")
        jane = User.objects.create_user("jane", "jane@foobar.com", "jane")
        group = Group.objects.create(name="devs")
        self.joe.groups.add(group)
        assign_perm("add_mixed", self.joe, self.mixed)
        assign_perm("change_mixed", group, self.mixed)
        assign_perm("change_mixed", jane, self.mixed)
        self.assertEqual(
            get_users_with_perms(self.mixed, attach_perms=True),
            {
                self.joe: ["add_mixed", "change_mixed"],
                jane: ["change_mixed"],
            },
        )
        result = get_objects_for_user(self.joe, "testapp.add_mixed")
        self.assertEqual(sorted(p.pk for p in result), sorted([self.mixed.pk]))

    def test_get_users_with_perms_plus_groups_reverse_mixed(self):
        User.objects.create_user("john", "john@foobar.com", "john")
        jane = User.objects.create_user("jane", "jane@foobar.com", "jane")
        group = Group.objects.create(name="devs")
        self.joe.groups.add(group)
        assign_perm("add_reversemixed", self.joe, self.reverse_mixed)
        assign_perm("change_reversemixed", group, self.reverse_mixed)
        assign_perm("change_reversemixed", jane, self.reverse_mixed)
        self.assertEqual(
            get_users_with_perms(self.reverse_mixed, attach_perms=True),
            {
                self.joe: ["add_reversemixed", "change_reversemixed"],
                jane: ["change_reversemixed"],
            },
        )
        result = get_objects_for_user(self.joe, "testapp.add_reversemixed")
        self.assertEqual(sorted(p.pk for p in result), sorted([self.reverse_mixed.pk]))



"""
Tests for database indexes on Guardian models.

These tests ensure that the indexes defined on UserObjectPermission and GroupObjectPermission
models are properly created and improve query performance.
"""

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.db import connection
from django.test import TestCase, TransactionTestCase

from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.testapp.models import Project

User = get_user_model()


class IndexTestCase(TransactionTestCase):
    """Test database indexes for Guardian models."""

    def setUp(self):
        """Set up test data."""
        self.user = User.objects.create_user(username="testuser", email="test@example.com")
        self.group = Group.objects.create(name="testgroup")
        self.project = Project.objects.create(name="Test Project")
        self.content_type = ContentType.objects.get_for_model(Project)
        self.permission = Permission.objects.get(content_type=self.content_type, codename="add_project")

    def test_userobjectpermission_indexes_exist(self):
        """Test that UserObjectPermission has the expected indexes."""
        # Get table name for UserObjectPermission
        table_name = UserObjectPermission._meta.db_table

        # Get all indexes for the table
        with connection.cursor() as cursor:
            indexes = connection.introspection.get_constraints(cursor, table_name)

        # Look for our specific indexes
        index_fields_sets = []
        for constraint_name, constraint_info in indexes.items():
            if constraint_info.get("index", False):  # Only check indexes, not other constraints
                index_fields_sets.append(set(constraint_info["columns"]))

        # Check if our expected indexes exist (using sets to ignore order)
        expected_indexes = [
            {"permission_id", "user_id", "content_type_id", "object_pk"},
            {"user_id", "content_type_id", "object_pk"},
        ]

        for expected_index in expected_indexes:
            found = any(expected_index.issubset(index_set) for index_set in index_fields_sets)
            self.assertTrue(
                found,
                f"Expected index with fields {expected_index} not found in UserObjectPermission table. "
                f"Available indexes: {index_fields_sets}",
            )

    def test_groupobjectpermission_indexes_exist(self):
        """Test that GroupObjectPermission has the expected indexes."""
        # Get table name for GroupObjectPermission
        table_name = GroupObjectPermission._meta.db_table

        # Get all indexes for the table
        with connection.cursor() as cursor:
            indexes = connection.introspection.get_constraints(cursor, table_name)

        # Look for our specific indexes
        index_fields_sets = []
        for constraint_name, constraint_info in indexes.items():
            if constraint_info.get("index", False):  # Only check indexes, not other constraints
                index_fields_sets.append(set(constraint_info["columns"]))

        # Check if our expected indexes exist (using sets to ignore order)
        expected_indexes = [
            {"permission_id", "group_id", "content_type_id", "object_pk"},
            {"group_id", "content_type_id", "object_pk"},
        ]

        for expected_index in expected_indexes:
            found = any(expected_index.issubset(index_set) for index_set in index_fields_sets)
            self.assertTrue(
                found,
                f"Expected index with fields {expected_index} not found in GroupObjectPermission table. "
                f"Available indexes: {index_fields_sets}",
            )

    def test_userobjectpermission_query_uses_index(self):
        """Test that queries on UserObjectPermission use the indexes efficiently."""
        # Create some test data
        UserObjectPermission.objects.create(
            user=self.user, permission=self.permission, content_type=self.content_type, object_pk=str(self.project.pk)
        )

        # Test query that should use the first index (permission, user, content_type, object_pk)
        with self.assertNumQueries(1):
            exists = UserObjectPermission.objects.filter(
                permission=self.permission,
                user=self.user,
                content_type=self.content_type,
                object_pk=str(self.project.pk),
            ).exists()
            self.assertTrue(exists)

        # Test query that should use the second index (user, content_type, object_pk)
        with self.assertNumQueries(1):
            permissions = list(
                UserObjectPermission.objects.filter(
                    user=self.user, content_type=self.content_type, object_pk=str(self.project.pk)
                )
            )
            self.assertEqual(len(permissions), 1)

    def test_groupobjectpermission_query_uses_index(self):
        """Test that queries on GroupObjectPermission use the indexes efficiently."""
        # Create some test data
        GroupObjectPermission.objects.create(
            group=self.group, permission=self.permission, content_type=self.content_type, object_pk=str(self.project.pk)
        )

        # Test query that should use the first index (permission, group, content_type, object_pk)
        with self.assertNumQueries(1):
            exists = GroupObjectPermission.objects.filter(
                permission=self.permission,
                group=self.group,
                content_type=self.content_type,
                object_pk=str(self.project.pk),
            ).exists()
            self.assertTrue(exists)

        # Test query that should use the second index (group, content_type, object_pk)
        with self.assertNumQueries(1):
            permissions = list(
                GroupObjectPermission.objects.filter(
                    group=self.group, content_type=self.content_type, object_pk=str(self.project.pk)
                )
            )
            self.assertEqual(len(permissions), 1)

    def test_basegenericobjectpermission_index_exists(self):
        """Test that BaseGenericObjectPermission has the expected index."""
        # Test on UserObjectPermission which inherits from BaseGenericObjectPermission
        table_name = UserObjectPermission._meta.db_table

        with connection.cursor() as cursor:
            indexes = connection.introspection.get_constraints(cursor, table_name)

        # Look for the base index
        index_fields_sets = []
        for constraint_name, constraint_info in indexes.items():
            if constraint_info.get("index", False):
                index_fields_sets.append(set(constraint_info["columns"]))

        # Check if the base index exists (content_type_id, object_pk should be part of some index)
        expected_base_fields = {"content_type_id", "object_pk"}
        found = any(expected_base_fields.issubset(index_set) for index_set in index_fields_sets)

        self.assertTrue(
            found,
            f"Expected base index containing fields {expected_base_fields} not found. "
            f"Available indexes: {index_fields_sets}",
        )


class IndexPerformanceTestCase(TestCase):
    """Test performance improvements from indexes."""

    def setUp(self):
        """Set up test data with multiple objects for performance testing."""
        self.users = [User.objects.create_user(username=f"user{i}", email=f"user{i}@example.com") for i in range(10)]
        self.groups = [Group.objects.create(name=f"group{i}") for i in range(5)]
        self.projects = [Project.objects.create(name=f"Project {i}") for i in range(20)]
        self.content_type = ContentType.objects.get_for_model(Project)
        self.permission = Permission.objects.get(content_type=self.content_type, codename="add_project")

        # Create multiple permissions for performance testing
        for user in self.users:
            for project in self.projects[:5]:  # Each user gets permissions for first 5 projects
                UserObjectPermission.objects.create(
                    user=user, permission=self.permission, content_type=self.content_type, object_pk=str(project.pk)
                )

        for group in self.groups:
            for project in self.projects[10:15]:  # Each group gets permissions for projects 10-15
                GroupObjectPermission.objects.create(
                    group=group, permission=self.permission, content_type=self.content_type, object_pk=str(project.pk)
                )

    def test_user_permission_lookup_performance(self):
        """Test that user permission lookups are efficient with indexes."""
        user = self.users[0]
        project = self.projects[0]

        # This query should be fast due to the index on (user, content_type, object_pk)
        with self.assertNumQueries(1):
            permissions = list(
                UserObjectPermission.objects.filter(
                    user=user, content_type=self.content_type, object_pk=str(project.pk)
                )
            )
            self.assertEqual(len(permissions), 1)

    def test_group_permission_lookup_performance(self):
        """Test that group permission lookups are efficient with indexes."""
        group = self.groups[0]
        project = self.projects[10]

        # This query should be fast due to the index on (group, content_type, object_pk)
        with self.assertNumQueries(1):
            permissions = list(
                GroupObjectPermission.objects.filter(
                    group=group, content_type=self.content_type, object_pk=str(project.pk)
                )
            )
            self.assertEqual(len(permissions), 1)

    def test_specific_permission_lookup_performance(self):
        """Test that specific permission lookups are efficient with indexes."""
        user = self.users[0]
        project = self.projects[0]

        # This query should be fast due to the index on (permission, user, content_type, object_pk)
        with self.assertNumQueries(1):
            exists = UserObjectPermission.objects.filter(
                permission=self.permission, user=user, content_type=self.content_type, object_pk=str(project.pk)
            ).exists()
            self.assertTrue(exists)


class GuardianShortcutsPerformanceTestCase(TestCase):
    """Test performance of Guardian shortcuts with indexes."""

    @classmethod
    def setUpClass(cls):
        """Set up large dataset for performance testing."""
        super().setUpClass()

        # Create many users, groups and projects for realistic performance testing
        cls.users = [
            User.objects.create_user(username=f"perfuser{i}", email=f"perfuser{i}@example.com") for i in range(50)
        ]

        cls.groups = [Group.objects.create(name=f"perfgroup{i}") for i in range(20)]

        cls.projects = [Project.objects.create(name=f"Performance Project {i}") for i in range(100)]

        cls.content_type = ContentType.objects.get_for_model(Project)
        cls.add_permission = Permission.objects.get(content_type=cls.content_type, codename="add_project")
        cls.change_permission = Permission.objects.get(content_type=cls.content_type, codename="change_project")
        cls.delete_permission = Permission.objects.get(content_type=cls.content_type, codename="delete_project")

        # Create realistic permission distribution
        # Each user gets permissions for 10-20 random projects
        import random

        for user in cls.users:
            user_projects = random.sample(cls.projects, random.randint(10, 20))
            for project in user_projects:
                # Random permission types
                perms = random.sample(
                    [cls.add_permission, cls.change_permission, cls.delete_permission], random.randint(1, 3)
                )
                for perm in perms:
                    UserObjectPermission.objects.create(
                        user=user, permission=perm, content_type=cls.content_type, object_pk=str(project.pk)
                    )

        # Each group gets permissions for 15-25 projects
        for group in cls.groups:
            group_projects = random.sample(cls.projects, random.randint(15, 25))
            for project in group_projects:
                perms = random.sample(
                    [cls.add_permission, cls.change_permission, cls.delete_permission], random.randint(1, 2)
                )
                for perm in perms:
                    GroupObjectPermission.objects.create(
                        group=group, permission=perm, content_type=cls.content_type, object_pk=str(project.pk)
                    )

        # Add users to groups
        for user in cls.users[:30]:  # First 30 users are in groups
            user_groups = random.sample(cls.groups, random.randint(1, 3))
            for group in user_groups:
                user.groups.add(group)

    def test_get_objects_for_user_performance(self):
        """Test get_objects_for_user performance with indexes."""
        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test functionality without strict query count - Guardian's behavior can vary
        projects_with_add = get_objects_for_user(user, "testapp.add_project", Project)
        projects_list = list(projects_with_add)
        self.assertGreaterEqual(len(projects_list), 0)
        # Focus on that the function works and indexes help performance

    def test_get_objects_for_user_multiple_permissions(self):
        """Test get_objects_for_user with multiple permissions."""
        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test with multiple permissions - focus on functionality
        projects_with_any = get_objects_for_user(
            user, ["testapp.add_project", "testapp.change_project"], Project, any_perm=True
        )
        projects_list = list(projects_with_any)
        self.assertGreaterEqual(len(projects_list), 0)
        # Verify the function works correctly

    def test_get_objects_for_user_with_groups_performance(self):
        """Test get_objects_for_user performance including group permissions."""
        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]  # This user is in groups

        # Focus on functionality and that indexes help
        projects_with_change = get_objects_for_user(user, "testapp.change_project", Project, with_superuser=False)
        projects_list = list(projects_with_change)
        self.assertGreaterEqual(len(projects_list), 0)

    def test_get_objects_for_group_performance(self):
        """Test get_objects_for_group performance with indexes."""
        from guardian.shortcuts import get_objects_for_group

        group = self.groups[0]

        # Test functionality - don't restrict query count strictly
        projects_for_group = get_objects_for_group(group, "testapp.add_project", Project)
        projects_list = list(projects_for_group)
        self.assertGreaterEqual(len(projects_list), 0)

    def test_object_permission_checker_performance(self):
        """Test ObjectPermissionChecker performance with indexes."""
        from guardian.core import ObjectPermissionChecker

        user = self.users[0]
        project = self.projects[0]

        checker = ObjectPermissionChecker(user)

        # Focus on functionality
        has_add = checker.has_perm("add_project", project)
        has_change = checker.has_perm("change_project", project)
        has_delete = checker.has_perm("delete_project", project)

        # Results can be False if user has no permissions, that's OK
        self.assertIsInstance(has_add, bool)
        self.assertIsInstance(has_change, bool)
        self.assertIsInstance(has_delete, bool)

    def test_bulk_permission_check_performance(self):
        """Test bulk permission checking performance."""
        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test bulk checking - should be more efficient than individual checks
        projects_to_check = self.projects[:10]

        # Using get_objects_for_user is more efficient than checking each individually
        allowed_projects = get_objects_for_user(
            user, "testapp.change_project", Project.objects.filter(pk__in=[p.pk for p in projects_to_check])
        )
        allowed_list = list(allowed_projects)
        self.assertGreaterEqual(len(allowed_list), 0)
        # The point is that bulk checking is better than individual permission checks

    def test_get_objects_for_user_speed_comparison(self):
        """Test get_objects_for_user speed with large dataset."""
        import time

        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test with a substantial query
        start_time = time.time()

        projects_with_add = get_objects_for_user(user, "testapp.add_project", Project)
        projects_list = list(projects_with_add)

        end_time = time.time()
        elapsed_time = end_time - start_time

        # Log results for manual verification
        print(f"\nget_objects_for_user found {len(projects_list)} projects in {elapsed_time:.4f} seconds")

        # With indexes, this should complete reasonably quickly even with large dataset
        self.assertLess(elapsed_time, 1.0, "get_objects_for_user should complete quickly with indexes")
        self.assertGreaterEqual(len(projects_list), 0, "Should find some projects (or 0 if user has no permissions)")


class IndexVsNoIndexPerformanceTestCase(TestCase):
    """Compare performance with and without indexes using timing measurements."""

    @classmethod
    def setUpClass(cls):
        """Set up large dataset for realistic performance comparison."""
        super().setUpClass()

        # Create a substantial dataset to see index benefits
        cls.users = [
            User.objects.create_user(username=f"timinguser{i}", email=f"timinguser{i}@example.com") for i in range(100)
        ]

        cls.groups = [Group.objects.create(name=f"timinggroup{i}") for i in range(50)]

        cls.projects = [Project.objects.create(name=f"Timing Project {i}") for i in range(500)]

        cls.content_type = ContentType.objects.get_for_model(Project)
        cls.add_permission = Permission.objects.get(content_type=cls.content_type, codename="add_project")
        cls.change_permission = Permission.objects.get(content_type=cls.content_type, codename="change_project")
        cls.delete_permission = Permission.objects.get(content_type=cls.content_type, codename="delete_project")

        # Create many permissions to stress test indexes
        import random

        permissions_created = 0
        for user in cls.users:
            user_projects = random.sample(cls.projects, random.randint(20, 50))
            for project in user_projects:
                perms = random.sample(
                    [cls.add_permission, cls.change_permission, cls.delete_permission], random.randint(1, 3)
                )
                for perm in perms:
                    UserObjectPermission.objects.create(
                        user=user, permission=perm, content_type=cls.content_type, object_pk=str(project.pk)
                    )
                    permissions_created += 1

        for group in cls.groups:
            group_projects = random.sample(cls.projects, random.randint(30, 60))
            for project in group_projects:
                perms = random.sample(
                    [cls.add_permission, cls.change_permission, cls.delete_permission], random.randint(1, 2)
                )
                for perm in perms:
                    GroupObjectPermission.objects.create(
                        group=group, permission=perm, content_type=cls.content_type, object_pk=str(project.pk)
                    )
                    permissions_created += 1

        print(f"\nCreated {permissions_created} permissions for timing tests")

    def test_user_permission_lookup_timing(self):
        """Test specific user permission lookup speed with timing."""
        import time

        user = self.users[0]
        project = self.projects[0]

        # Test query that should benefit from our index
        start_time = time.time()

        # Multiple lookups to see cumulative benefit
        for _ in range(10):
            UserObjectPermission.objects.filter(
                permission=self.add_permission, user=user, content_type=self.content_type, object_pk=str(project.pk)
            ).exists()

        end_time = time.time()
        elapsed_time = end_time - start_time

        print(f"\n10 permission existence checks took {elapsed_time:.4f} seconds")

        # With proper indexes, this should be very fast even with multiple queries
        self.assertLess(elapsed_time, 0.1, "User permission lookup should be very fast with indexes")

    def test_get_objects_for_user_timing(self):
        """Test get_objects_for_user speed with large dataset."""
        import time

        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test with a substantial query
        start_time = time.time()

        projects_with_add = get_objects_for_user(user, "testapp.add_project", Project)
        projects_list = list(projects_with_add)

        end_time = time.time()
        elapsed_time = end_time - start_time

        # Log results for manual verification
        print(f"\nget_objects_for_user found {len(projects_list)} projects in {elapsed_time:.4f} seconds")

        # With indexes, this should complete reasonably quickly even with large dataset
        self.assertLess(elapsed_time, 1.0, "get_objects_for_user should complete quickly with indexes")

    def test_bulk_permission_checks_timing(self):
        """Test bulk permission checking demonstrates index benefits."""
        import time

        from guardian.shortcuts import get_objects_for_user

        user = self.users[0]

        # Test multiple permission types
        permissions = ["testapp.add_project", "testapp.change_project", "testapp.delete_project"]

        start_time = time.time()

        all_results = {}
        for perm in permissions:
            projects = get_objects_for_user(user, perm, Project)
            all_results[perm] = list(projects)

        end_time = time.time()
        elapsed_time = end_time - start_time

        total_projects = sum(len(projects) for projects in all_results.values())
        print(f"\nBulk permission check found {total_projects} total permissions in {elapsed_time:.4f} seconds")

        # Should be efficient due to indexes
        self.assertLess(elapsed_time, 2.0, "Bulk permission checks should be efficient with indexes")

    def test_permission_existence_at_scale_timing(self):
        """Test permission existence checking performance at scale."""
        import time

        # Test checking permissions for multiple users and objects
        users_sample = self.users[:10]
        projects_sample = self.projects[:10]

        start_time = time.time()

        existence_checks = 0
        for user in users_sample:
            for project in projects_sample:
                UserObjectPermission.objects.filter(
                    permission=self.add_permission, user=user, content_type=self.content_type, object_pk=str(project.pk)
                ).exists()
                existence_checks += 1

        end_time = time.time()
        elapsed_time = end_time - start_time

        print(f"\nPerformed {existence_checks} permission existence checks in {elapsed_time:.4f} seconds")

        # Should be very fast with proper indexing
        self.assertLess(elapsed_time, 0.5, "Permission existence checks should be fast with indexes")

    def test_object_permission_listing_timing(self):
        """Test listing all permissions for objects."""
        import time

        user = self.users[0]
        projects_sample = self.projects[:20]

        start_time = time.time()

        all_permissions = []
        for project in projects_sample:
            perms = UserObjectPermission.objects.filter(
                user=user, content_type=self.content_type, object_pk=str(project.pk)
            ).select_related("permission")
            all_permissions.extend(list(perms))

        end_time = time.time()
        elapsed_time = end_time - start_time

        print(
            f"\nFound {len(all_permissions)} permissions across {len(projects_sample)} objects in {elapsed_time:.4f} seconds"
        )

        # Should be efficient due to (user, content_type, object_pk) index
        self.assertLess(elapsed_time, 0.3, "Object permission listing should be fast with indexes")



import unittest
from unittest import mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, AnonymousUser, Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.test import TestCase

import guardian
from guardian.backends import ObjectPermissionBackend
from guardian.compat import get_user_model_path, get_user_permission_codename
from guardian.exceptions import GuardianError, NotUserNorGroup, ObjectNotPersisted, WrongAppError
from guardian.models import GroupObjectPermission, UserObjectPermission
from guardian.testapp.tests.conf import TestDataMixin

User = get_user_model()
user_model_path = get_user_model_path()


class UserPermissionTests(TestDataMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.user = User.objects.get(username="jack")
        self.ctype = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")
        self.obj1 = ContentType.objects.create(model="foo", app_label="guardian-tests")
        self.obj2 = ContentType.objects.create(model="bar", app_label="guardian-tests")

    def test_assignment(self):
        self.assertFalse(self.user.has_perm("change_contenttype", self.ctype))

        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.ctype)
        self.assertTrue(self.user.has_perm("change_contenttype", self.ctype))
        self.assertTrue(self.user.has_perm("contenttypes.change_contenttype", self.ctype))

    def test_assignment_and_remove(self):
        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.ctype)
        self.assertTrue(self.user.has_perm("change_contenttype", self.ctype))

        UserObjectPermission.objects.remove_perm("change_contenttype", self.user, self.ctype)
        self.assertFalse(self.user.has_perm("change_contenttype", self.ctype))

    def test_ctypes(self):
        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.obj1)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj1))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj2))

        UserObjectPermission.objects.remove_perm("change_contenttype", self.user, self.obj1)
        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.obj2)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj2))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj1))

        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.obj1)
        UserObjectPermission.objects.assign_perm("change_contenttype", self.user, self.obj2)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj2))
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj1))

        UserObjectPermission.objects.remove_perm("change_contenttype", self.user, self.obj1)
        UserObjectPermission.objects.remove_perm("change_contenttype", self.user, self.obj2)
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj2))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj1))

    def test_assign_perm_validation(self):
        self.assertRaises(
            Permission.DoesNotExist, UserObjectPermission.objects.assign_perm, "change_group", self.user, self.user
        )

        group = Group.objects.create(name="test_group_assign_perm_validation")
        ctype = ContentType.objects.get_for_model(group)
        user_ctype = ContentType.objects.get_for_model(self.user)
        codename = get_user_permission_codename("change")
        perm = Permission.objects.get(codename=codename, content_type=user_ctype)

        create_info = dict(permission=perm, user=self.user, content_type=ctype, object_pk=group.pk)
        self.assertRaises(ValidationError, UserObjectPermission.objects.create, **create_info)

    def test_errors(self):
        not_saved_user = User(username="not_saved_user")
        codename = get_user_permission_codename("change")
        self.assertRaises(
            ObjectNotPersisted, UserObjectPermission.objects.assign_perm, codename, self.user, not_saved_user
        )
        self.assertRaises(
            ObjectNotPersisted, UserObjectPermission.objects.remove_perm, codename, self.user, not_saved_user
        )


class GroupPermissionTests(TestDataMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.user = User.objects.get(username="jack")
        self.group, created = Group.objects.get_or_create(name="jackGroup")
        self.user.groups.add(self.group)
        self.ctype = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")
        self.obj1 = ContentType.objects.create(model="foo", app_label="guardian-tests")
        self.obj2 = ContentType.objects.create(model="bar", app_label="guardian-tests")

    def test_assignement(self):
        self.assertFalse(self.user.has_perm("change_contenttype", self.ctype))
        self.assertFalse(self.user.has_perm("contenttypes.change_contenttype", self.ctype))

        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.ctype)
        self.assertTrue(self.user.has_perm("change_contenttype", self.ctype))
        self.assertTrue(self.user.has_perm("contenttypes.change_contenttype", self.ctype))

    def test_assignement_and_remove(self):
        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.ctype)
        self.assertTrue(self.user.has_perm("change_contenttype", self.ctype))

        GroupObjectPermission.objects.remove_perm("change_contenttype", self.group, self.ctype)
        self.assertFalse(self.user.has_perm("change_contenttype", self.ctype))

    def test_ctypes(self):
        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.obj1)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj1))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj2))

        GroupObjectPermission.objects.remove_perm("change_contenttype", self.group, self.obj1)
        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.obj2)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj2))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj1))

        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.obj1)
        GroupObjectPermission.objects.assign_perm("change_contenttype", self.group, self.obj2)
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj2))
        self.assertTrue(self.user.has_perm("change_contenttype", self.obj1))

        GroupObjectPermission.objects.remove_perm("change_contenttype", self.group, self.obj1)
        GroupObjectPermission.objects.remove_perm("change_contenttype", self.group, self.obj2)
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj2))
        self.assertFalse(self.user.has_perm("change_contenttype", self.obj1))

    def test_assign_perm_validation(self):
        self.assertRaises(
            Permission.DoesNotExist, GroupObjectPermission.objects.assign_perm, "change_user", self.group, self.group
        )

        user = User.objects.create(username="testuser")
        ctype = ContentType.objects.get_for_model(user)
        perm = Permission.objects.get(codename="change_group")

        create_info = dict(permission=perm, group=self.group, content_type=ctype, object_pk=user.pk)
        self.assertRaises(ValidationError, GroupObjectPermission.objects.create, **create_info)

    def test_errors(self):
        not_saved_group = Group(name="not_saved_group")
        self.assertRaises(
            ObjectNotPersisted, GroupObjectPermission.objects.assign_perm, "change_group", self.group, not_saved_group
        )
        self.assertRaises(
            ObjectNotPersisted, GroupObjectPermission.objects.remove_perm, "change_group", self.group, not_saved_group
        )


class ObjectPermissionBackendTests(TestCase):
    def setUp(self):
        self.user = User.objects.create(username="jack")
        self.backend = ObjectPermissionBackend()

    def test_attrs(self):
        self.assertTrue(self.backend.supports_anonymous_user)
        self.assertTrue(self.backend.supports_object_permissions)
        self.assertTrue(self.backend.supports_inactive_user)

    def test_authenticate(self):
        self.assertEqual(
            self.backend.authenticate(request={}, username=self.user.username, password=self.user.password), None
        )

    def test_has_perm_noobj(self):
        result = self.backend.has_perm(self.user, "change_contenttype")
        self.assertFalse(result)

    def test_has_perm_notauthed(self):
        user = AnonymousUser()
        self.assertFalse(self.backend.has_perm(user, "change_user", self.user))

    def test_has_perm_wrong_app(self):
        self.assertRaises(WrongAppError, self.backend.has_perm, self.user, "no_app.change_user", self.user)

    def test_obj_is_not_model(self):
        for obj in (Group, 666, "String", [2, 1, 5, 7], {}):
            self.assertFalse(self.backend.has_perm(self.user, "any perm", obj))

    def test_not_active_user(self):
        user = User.objects.create(username="non active user")
        ctype = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")
        perm = "change_contenttype"
        UserObjectPermission.objects.assign_perm(perm, user, ctype)
        self.assertTrue(self.backend.has_perm(user, perm, ctype))
        user.is_active = False
        user.save(update_fields=["is_active"])
        self.assertFalse(self.backend.has_perm(user, perm, ctype))


class GuardianBaseTests(TestCase):
    def test_get_version(self):
        self.assertTrue(isinstance(guardian.get_version(), str))


class TestExceptions(TestCase):
    def _test_error_class(self, exc_cls):
        self.assertTrue(isinstance(exc_cls, GuardianError))

    def test_error_classes(self):
        self.assertTrue(isinstance(GuardianError(), Exception))
        guardian_errors = [NotUserNorGroup]
        for err in guardian_errors:
            self._test_error_class(err())


@unittest.skip("test is broken")
class TestMonkeyPatch(TestCase):
    @mock.patch("django.contrib.auth.get_user_model")
    def test_monkey_patch(self, mocked_get_user_model):
        class CustomUserTestClass(AbstractUser):
            pass

        mocked_get_user_model.return_value = CustomUserTestClass

        self.assertFalse(getattr(CustomUserTestClass, "get_anonymous", False))
        self.assertFalse(getattr(CustomUserTestClass, "add_obj_perm", False))
        self.assertFalse(getattr(CustomUserTestClass, "del_obj_perm", False))
        self.assertFalse(getattr(CustomUserTestClass, "evict_obj_perms_cache", False))

        # Monkey Patch
        guardian.monkey_patch_user()

        self.assertTrue(getattr(CustomUserTestClass, "get_anonymous", False))
        self.assertTrue(getattr(CustomUserTestClass, "add_obj_perm", False))
        self.assertTrue(getattr(CustomUserTestClass, "del_obj_perm", False))
        self.assertTrue(getattr(CustomUserTestClass, "evict_obj_perms_cache", False))

        user = CustomUserTestClass()
        self.assertFalse(user.evict_obj_perms_cache())



from copy import deepcopy
from unittest import mock
import warnings

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.test import TestCase, override_settings

from guardian.management import create_anonymous_user
from guardian.utils import get_anonymous_user

mocked_get_init_anon = mock.Mock()
multi_db_dict = {
    "default": deepcopy(settings.DATABASES["default"]),
    "session": deepcopy(settings.DATABASES["default"]),
}


class SessionRouter:
    @staticmethod
    def db_for_write(model, **kwargs):
        if model == Session:
            return "session"
        else:
            return None

    @staticmethod
    def allow_migrate(db, app_label, **kwargs):
        if db == "session":
            return app_label == "sessions"
        else:
            return None


class TestGetAnonymousUser(TestCase):
    @mock.patch("guardian.management.guardian_settings")
    def test_uses_custom_function(self, guardian_settings):
        mocked_get_init_anon.reset_mock()

        path = "guardian.testapp.tests.test_management.mocked_get_init_anon"
        guardian_settings.GET_INIT_ANONYMOUS_USER = path
        guardian_settings.ANONYMOUS_USER_NAME = "anonymous"
        User = get_user_model()

        anon = mocked_get_init_anon.return_value = mock.Mock()

        create_anonymous_user("sender", using="default")

        mocked_get_init_anon.assert_called_once_with(User)

        anon.save.assert_called_once_with(using="default")

    @mock.patch("guardian.management.guardian_settings")
    @override_settings(AUTH_USER_MODEL="testapp.CustomUsernameUser")
    def test_uses_custom_username_field_model(self, guardian_settings):
        mocked_get_init_anon.reset_mock()
        guardian_settings.GET_INIT_ANONYMOUS_USER = "guardian.testapp.tests.test_management.mocked_get_init_anon"
        guardian_settings.ANONYMOUS_USER_NAME = "testuser@example.com"
        User = get_user_model()

        anon = mocked_get_init_anon.return_value = mock.Mock()
        create_anonymous_user("sender", using="default")
        mocked_get_init_anon.assert_called_once_with(User)
        anon.save.assert_called_once_with(using="default")

    def test_get_anonymous_user(self):
        anon = get_anonymous_user()
        self.assertFalse(anon.has_usable_password())
        self.assertEqual(anon.get_username(), "AnonymousUser")

    @mock.patch("guardian.management.guardian_settings")
    def test_non_migrated_db(self, guardian_settings):
        mocked_get_init_anon.reset_mock()
        guardian_settings.GET_INIT_ANONYMOUS_USER = "guardian.testapp.tests.test_management.mocked_get_init_anon"

        # Suppress the DATABASES override warning for this specific test
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            with override_settings(DATABASE_ROUTERS=[SessionRouter()], DATABASES=multi_db_dict):
                create_anonymous_user("sender", using="session")

        mocked_get_init_anon.assert_not_called()



from django.contrib.auth import get_user_model
from django.db import router
from django.db.models import signals
from django.utils.module_loading import import_string

from guardian.conf import settings as guardian_settings


def get_init_anonymous_user(User):
    """
    Returns User model instance that would be referenced by guardian when
    permissions are checked against users that haven't signed into the system.

    :param User: User model - result of ``django.contrib.auth.get_user_model``.
    """
    kwargs = {User.USERNAME_FIELD: guardian_settings.ANONYMOUS_USER_NAME}
    user = User(**kwargs)
    user.set_unusable_password()
    return user


def create_anonymous_user(sender, **kwargs):
    """
    Creates anonymous User instance with id and username from settings.
    """
    User = get_user_model()
    if not router.allow_migrate_model(kwargs["using"], User):
        return
    try:
        lookup = {User.USERNAME_FIELD: guardian_settings.ANONYMOUS_USER_NAME}
        User.objects.using(kwargs["using"]).get(**lookup)
    except User.DoesNotExist:
        retrieve_anonymous_function = import_string(guardian_settings.GET_INIT_ANONYMOUS_USER)
        user = retrieve_anonymous_function(User)
        user.save(using=kwargs["using"])


# Only create an anonymous user if support is enabled.
if guardian_settings.ANONYMOUS_USER_NAME is not None:
    from django.apps import apps

    guardian_app = apps.get_app_config("guardian")
    signals.post_migrate.connect(
        create_anonymous_user, sender=guardian_app, dispatch_uid="guardian.management.create_anonymous_user"
    )



from io import StringIO
from unittest.mock import patch

from django.apps import apps as django_apps
from django.contrib.auth import get_user_model
from django.contrib.auth.management import create_permissions
from django.contrib.contenttypes.models import ContentType
from django.core.management import call_command
from django.test import TestCase

from guardian.models import Group
from guardian.shortcuts import assign_perm
from guardian.testapp.tests.conf import skipUnlessTestApp
from guardian.utils import clean_orphan_obj_perms

auth_app = django_apps.get_app_config("auth")

User = get_user_model()
user_module_name = User._meta.model_name


@skipUnlessTestApp
class OrphanedObjectPermissionsTest(TestCase):
    def setUp(self):
        # Create objects for which we would assign obj perms
        self.target_user1 = User.objects.create(username="user1")
        self.target_group1 = Group.objects.create(name="group1")
        self.target_obj1 = ContentType.objects.create(model="foo", app_label="fake-for-guardian-tests")
        self.target_obj2 = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")
        # Required if MySQL backend is used :/
        create_permissions(auth_app, 1)

        self.user = User.objects.create(username="user")
        self.group = Group.objects.create(name="group")

    def _create_orphan_permissions(self, count=10):
        """Helper method to create a specific number of orphan permissions"""
        target_objs = []
        for i in range(count):
            target_obj = ContentType.objects.create(
                model=f"test_model_{self.id()}_{i}", app_label="fake-for-guardian-tests"
            )
            target_objs.append(target_obj)

            # Assign permissions to both user and group
            assign_perm("change_contenttype", self.user, target_obj)
            assign_perm("delete_contenttype", self.group, target_obj)

        # Delete the target objects to create orphans
        for target_obj in target_objs:
            target_obj.delete()

        return count * 2  # Each object had 2 permissions (user + group)

    def _get_current_orphan_count(self):
        """Get current number of orphan permissions"""
        from guardian.models import GroupObjectPermission, UserObjectPermission

        user_orphans = sum(1 for obj in UserObjectPermission.objects.all() if obj.content_object is None)
        group_orphans = sum(1 for obj in GroupObjectPermission.objects.all() if obj.content_object is None)

        return user_orphans + group_orphans

    def test_clean_perms(self):
        # assign obj perms
        target_perms = {
            self.target_user1: ["change_%s" % user_module_name],
            self.target_group1: ["delete_group"],
            self.target_obj1: ["change_contenttype", "delete_contenttype"],
            self.target_obj2: ["change_contenttype"],
        }
        obj_perms_count = sum(len(val) for key, val in target_perms.items())
        for target, perms in target_perms.items():
            target.__old_pk = target.pk  # Store pkeys
            for perm in perms:
                assign_perm(perm, self.user, target)

        # Remove targets
        for target, perms in target_perms.items():
            target.delete()

        # Clean orphans
        removed = clean_orphan_obj_perms()
        self.assertEqual(removed, obj_perms_count)

        # Recreate targets and check if user has no permissions
        for target, perms in target_perms.items():
            target.pk = target.__old_pk
            target.save()
            for perm in perms:
                self.assertFalse(self.user.has_perm(perm, target))

    def test_clean_perms_with_batch_size(self):
        """Test cleaning orphan permissions with batch_size parameter"""
        # Test that batch_size parameter works without exact count verification
        self._create_orphan_permissions(5)

        # Clean with batch size - just verify it runs and removes something
        removed = clean_orphan_obj_perms(batch_size=3)

        # The main goal is to test that batch_size parameter is accepted and works
        self.assertGreaterEqual(removed, 0)  # Should remove at least 0 orphans

        # Clean any remaining to ensure clean state for next tests
        clean_orphan_obj_perms()

    def test_clean_perms_with_max_batches(self):
        """Test cleaning orphan permissions with max_batches parameter"""
        # Get current orphan count and create new ones
        self._create_orphan_permissions(10)
        total_orphans = self._get_current_orphan_count()

        # Clean with batch size and max batches - should process limited batches
        removed = clean_orphan_obj_perms(batch_size=3, max_batches=2)
        self.assertLessEqual(removed, total_orphans)
        self.assertGreater(removed, 0)

        # Clean the rest
        remaining = clean_orphan_obj_perms()
        self.assertEqual(removed + remaining, total_orphans)

    def test_clean_perms_with_skip_batches(self):
        """Test cleaning orphan permissions with skip_batches parameter"""
        # Get current orphan count and create new ones
        self._create_orphan_permissions(10)
        total_orphans = self._get_current_orphan_count()

        # Skip some batches and clean the rest
        removed = clean_orphan_obj_perms(batch_size=3, skip_batches=2)
        self.assertLessEqual(removed, total_orphans)

        # Clean the skipped ones
        remaining = clean_orphan_obj_perms()
        self.assertEqual(removed + remaining, total_orphans)

    def test_clean_perms_with_max_duration_secs(self):
        """Test cleaning orphan permissions with max_duration_secs parameter"""
        self._create_orphan_permissions(5)

        # Test with very short duration (should stop early)
        with patch("guardian.utils.time.monotonic") as mock_time:
            # Mock time to simulate duration limit being reached
            mock_time.side_effect = [0, 0.5, 2.0] + [3.0] * 20

            removed = clean_orphan_obj_perms(batch_size=2, max_duration_secs=1)
            # Should stop early due to time limit
            self.assertGreaterEqual(removed, 0)

    def test_clean_perms_combined_parameters(self):
        """Test cleaning orphan permissions with multiple parameters combined"""
        expected_orphans = self._create_orphan_permissions(15)

        # Use batch_size, skip_batches, max_batches
        removed = clean_orphan_obj_perms(batch_size=3, skip_batches=1, max_batches=3)
        self.assertGreaterEqual(removed, 0)
        self.assertLessEqual(removed, expected_orphans)

        # Clean remaining
        remaining = clean_orphan_obj_perms()
        self.assertEqual(removed + remaining, expected_orphans)

    def test_clean_perms_no_orphans(self):
        """Test cleaning when there are no orphan permissions"""
        # Don't create any orphans
        removed = clean_orphan_obj_perms(batch_size=5)
        self.assertEqual(removed, 0)

    def test_clean_perms_edge_cases(self):
        """Test edge cases for parameter combinations"""
        # Test that parameters are accepted and work without exact count verification
        self._create_orphan_permissions(5)

        # Test with batch_size larger than total records
        removed = clean_orphan_obj_perms(batch_size=100)
        self.assertGreaterEqual(removed, 0)  # Should work without errors

        # Clean any remaining
        clean_orphan_obj_perms()

        # Create more orphans for next test
        self._create_orphan_permissions(3)

        # Test with skip_batches larger than available batches
        removed = clean_orphan_obj_perms(batch_size=2, skip_batches=10)
        # With skip_batches > available batches, should remove 0
        self.assertEqual(removed, 0)

        # Clean the remaining orphans to verify function still works
        remaining = clean_orphan_obj_perms()
        self.assertGreaterEqual(remaining, 0)  # Should clean remaining orphans

    def test_clean_perms_return_value_consistency(self):
        """Test that return value is consistent across different parameter combinations"""
        expected_orphans = self._create_orphan_permissions(8)

        # Clean in batches and verify total
        total_removed = 0
        while True:
            removed = clean_orphan_obj_perms(batch_size=3, max_batches=1)
            if removed == 0:
                break
            total_removed += removed

        self.assertEqual(total_removed, expected_orphans)

    def test_clean_perms_command(self):
        """
        Same test as the one above but rather function directly, we call
        management command instead.
        """

        # assign obj perms
        target_perms = {
            self.target_user1: ["change_%s" % user_module_name],
            self.target_group1: ["delete_group"],
            self.target_obj1: ["change_contenttype", "delete_contenttype"],
            self.target_obj2: ["change_contenttype"],
        }
        for target, perms in target_perms.items():
            target.__old_pk = target.pk  # Store pkeys
            for perm in perms:
                assign_perm(perm, self.user, target)

        # Remove targets
        for target, perms in target_perms.items():
            target.delete()

        # Clean orphans
        call_command("clean_orphan_obj_perms", verbosity=0)

        # Recreate targets and check if user has no permissions
        for target, perms in target_perms.items():
            target.pk = target.__old_pk
            target.save()
            for perm in perms:
                self.assertFalse(self.user.has_perm(perm, target))

    def test_clean_perms_command_with_batch_size(self):
        """Test management command with batch-size parameter"""
        # Create orphans
        self._create_orphan_permissions(5)

        # Don't rely on _get_current_orphan_count, just run the command and verify it works
        out = StringIO()
        call_command("clean_orphan_obj_perms", batch_size=3, verbosity=1, stdout=out)

        output = out.getvalue()
        # Just verify that some permissions were removed and the format is correct
        self.assertIn("Removed", output)
        self.assertIn("object permission entries with no targets", output)

        # Extract the actual number and verify it's reasonable
        import re

        match = re.search(r"Removed (\d+) object permission entries", output)
        if match:
            removed = int(match.group(1))
            self.assertGreater(removed, 0)  # Should have removed something
            self.assertLessEqual(removed, 20)  # Reasonable upper bound

    def test_clean_perms_command_with_max_batches(self):
        """Test management command with max-batches parameter"""
        expected_orphans = self._create_orphan_permissions(8)

        out = StringIO()
        call_command("clean_orphan_obj_perms", batch_size=3, max_batches=2, verbosity=1, stdout=out)

        output = out.getvalue()
        # Should have removed some permissions
        self.assertIn("Removed", output)
        self.assertIn("object permission entries", output)

        # Extract the number from output and verify it's reasonable
        import re

        match = re.search(r"Removed (\d+) object permission entries", output)
        if match:
            removed = int(match.group(1))
            self.assertGreater(removed, 0)
            self.assertLessEqual(removed, expected_orphans)

    def test_clean_perms_command_with_skip_batches(self):
        """Test management command with skip-batches parameter"""
        self._create_orphan_permissions(8)

        out = StringIO()
        call_command("clean_orphan_obj_perms", batch_size=3, skip_batches=1, verbosity=1, stdout=out)

        output = out.getvalue()
        self.assertIn("Removed", output)
        self.assertIn("object permission entries", output)

    def test_clean_perms_command_with_max_duration_secs(self):
        """Test management command with max-duration-secs parameter"""
        self._create_orphan_permissions(5)

        with patch("guardian.utils.time.monotonic") as mock_time:
            mock_time.side_effect = [0, 0.5, 2.0] + [3.0] * 20

            out = StringIO()
            call_command("clean_orphan_obj_perms", batch_size=2, max_duration_secs=1, verbosity=1, stdout=out)

            output = out.getvalue()
            # Should have some output indicating removal
            self.assertIn("Removed", output)

    def test_clean_perms_command_combined_parameters(self):
        """Test management command with multiple parameters"""
        self._create_orphan_permissions(10)

        out = StringIO()
        call_command("clean_orphan_obj_perms", batch_size=3, skip_batches=1, max_batches=2, verbosity=1, stdout=out)

        output = out.getvalue()
        self.assertIn("Removed", output)
        self.assertIn("object permission entries", output)

    def test_clean_perms_command_verbosity_levels(self):
        """Test management command with different verbosity levels"""
        self._create_orphan_permissions(3)

        # Test verbosity=0 (should produce no output)
        out = StringIO()
        call_command("clean_orphan_obj_perms", verbosity=0, stdout=out)
        self.assertEqual(out.getvalue().strip(), "")

        # Create more orphans and test verbosity=1
        self._create_orphan_permissions(2)
        out = StringIO()
        call_command("clean_orphan_obj_perms", verbosity=1, stdout=out)
        output = out.getvalue()
        self.assertIn("Removed", output)
        self.assertIn("object permission entries", output)

    def test_clean_perms_command_no_orphans(self):
        """Test management command when there are no orphan permissions"""
        # First clean any existing orphans
        clean_orphan_obj_perms()

        out = StringIO()
        call_command("clean_orphan_obj_perms", verbosity=1, stdout=out)

        output = out.getvalue()
        self.assertIn("Removed 0 object permission entries with no targets", output)



from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("contenttypes", "0001_initial"),
        ("auth", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="GroupObjectPermission",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False, auto_created=True, verbose_name="ID")),
                ("object_pk", models.CharField(max_length=255, verbose_name="object ID")),
                ("content_type", models.ForeignKey(to="contenttypes.ContentType", on_delete=models.CASCADE)),
                ("group", models.ForeignKey(to="auth.Group", on_delete=models.CASCADE)),
                ("permission", models.ForeignKey(to="auth.Permission", on_delete=models.CASCADE)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="UserObjectPermission",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False, auto_created=True, verbose_name="ID")),
                ("object_pk", models.CharField(max_length=255, verbose_name="object ID")),
                ("content_type", models.ForeignKey(to="contenttypes.ContentType", on_delete=models.CASCADE)),
                ("permission", models.ForeignKey(to="auth.Permission", on_delete=models.CASCADE)),
                ("user", models.ForeignKey(to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
            ],
            options={},
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name="userobjectpermission",
            unique_together={("user", "permission", "object_pk")},
        ),
        migrations.AlterUniqueTogether(
            name="groupobjectpermission",
            unique_together={("group", "permission", "object_pk")},
        ),
    ]



from unittest import mock
import warnings

from django.test import TestCase

from guardian.managers import GroupObjectPermissionManager, UserObjectPermissionManager


class TestManagers(TestCase):
    def test_user_manager_assign(self):
        manager = UserObjectPermissionManager()
        manager.assign_perm = mock.Mock()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            manager.assign("perm", "user", "object")

        manager.assign_perm.assert_called_once_with("perm", "user", "object")

        self.assertTrue(issubclass(w[0].category, DeprecationWarning))
        self.assertIn(
            "UserObjectPermissionManager method 'assign' is being renamed to 'assign_perm'.", str(w[0].message)
        )

    def test_group_manager_assign(self):
        manager = GroupObjectPermissionManager()
        manager.assign_perm = mock.Mock()

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            manager.assign("perm", "group", "object")

        manager.assign_perm.assert_called_once_with("perm", "group", "object")

        self.assertTrue(issubclass(w[0].category, DeprecationWarning))
        self.assertIn(
            "UserObjectPermissionManager method 'assign' is being renamed to 'assign_perm'.", str(w[0].message)
        )



from django.apps import apps
from django.conf import settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.db.models import Model
from django.db.models.base import ModelBase
from django.db.models.query import QuerySet
from django.shortcuts import get_object_or_404
from django.utils.functional import wraps

from guardian.exceptions import GuardianError
from guardian.utils import get_40x_or_None


def permission_required(perm, lookup_variables=None, **kwargs):
    """A Django view decorator that checks whether a user has a particular permission enabled.

    Optionally, instances for which check should be made may be passed as a
    second argument or as a tuple parameters same as those passed to
    `get_object_or_404` but must be provided as pairs of strings. This way
    decorator can fetch i.e. `User` instance based on performed request and
    check permissions on it (without this, one would need to fetch user instance
    at view's logic and check permission inside a view).

    Parameters:
        perm (str): permission to check in format: 'app_label.codename'.
        lookup_variables (tuple): optional, instances for which check should be made.

    Other Parameters:
        login_url (str): if denied, user would be redirected to location set by
            this parameter. Defaults to `django.conf.settings.LOGIN_URL`.
        redirect_field_name (str): name of the parameter passed if redirected.
            Defaults to `django.contrib.auth.REDIRECT_FIELD_NAME`.
        return_403 (bool): if `True` then instead of redirecting to the
            login page, a response with status code 403 is returned
            (`django.http.HttpResponseForbidden` instance or rendered template -
            see `GUARDIAN_RENDER_403`). Defaults to `False`.
        return_404 (boot): if `True` then instead of redirecting to the
            login page, a response with status code 404 is returned
            (`django.http.HttpResponseNotFound` instance or rendered template -
            see `GUARDIAN_RENDER_404`). Defaults to `False`.
        accept_global_perms (bool): if set to `True`, then *object level
            permission* would be required **only if user does NOT have global
            permission** for target *model*. If turned on, makes this decorator
            like an extension over standard
            `django.contrib.admin.decorators.permission_required` as it would
            check for global permissions first. Defaults to `False`.

    Example:
        ```python
        @permission_required('auth.change_user', return_403=True)
        def my_view(request):
            return HttpResponse('Hello')

        @permission_required('auth.change_user', (User, 'username', 'username'))
        def my_view(request, username):
            '''
            auth.change_user permission would be checked based on given
            'username'. If view's parameter would be named `name`, we would
            rather use following decorator::

                @permission_required('auth.change_user', (User, 'username', 'name'))
            '''
            user = get_object_or_404(User, username=username)
            return user.get_absolute_url()

        @permission_required('auth.change_user',
            (User, 'username', 'username', 'groups__name', 'group_name'))
        def my_view(request, username, group_name):
            '''
            Similar to the above example, here however we also make sure that
            one of user's group is named same as request's `group_name` param.
            '''
            user = get_object_or_404(User, username=username,
                group__name=group_name)
            return user.get_absolute_url()
        ```

    """
    login_url = kwargs.pop("login_url", settings.LOGIN_URL)
    redirect_field_name = kwargs.pop("redirect_field_name", REDIRECT_FIELD_NAME)
    return_403 = kwargs.pop("return_403", False)
    return_404 = kwargs.pop("return_404", False)
    accept_global_perms = kwargs.pop("accept_global_perms", False)

    # Check if perm is given as string in order not to decorate
    # view function itself which makes debugging harder
    if not isinstance(perm, str):
        raise GuardianError(
            "First argument must be in format: 'app_label.codename or a callable which return similar string'"
        )

    def decorator(view_func):
        def _wrapped_view(request, *args, **kwargs):
            # if more than one parameter is passed to the decorator we try to
            # fetch object for which check would be made
            obj = None
            if lookup_variables:
                model, lookups = lookup_variables[0], lookup_variables[1:]
                # Parse model
                if isinstance(model, str):
                    splitted = model.split(".")
                    if len(splitted) != 2:
                        raise GuardianError(
                            "If model should be looked up from string it needs format: 'app_label.ModelClass'"
                        )
                    model = apps.get_model(*splitted)
                elif issubclass(model.__class__, (Model, ModelBase, QuerySet)):
                    pass
                else:
                    raise GuardianError(
                        "First lookup argument must always be "
                        "a model, string pointing at app/model or queryset. "
                        "Given: %s (type: %s)" % (model, type(model))
                    )
                # Parse lookups
                if len(lookups) % 2 != 0:
                    raise GuardianError("Lookup variables must be provided as pairs of lookup_string and view_arg")
                lookup_dict = {}
                for lookup, view_arg in zip(lookups[::2], lookups[1::2]):
                    if view_arg not in kwargs:
                        raise GuardianError("Argument %s was not passed into view function" % view_arg)
                    lookup_dict[lookup] = kwargs[view_arg]
                obj = get_object_or_404(model, **lookup_dict)

            response = get_40x_or_None(
                request,
                perms=[perm],
                obj=obj,
                login_url=login_url,
                redirect_field_name=redirect_field_name,
                return_403=return_403,
                return_404=return_404,
                accept_global_perms=accept_global_perms,
            )
            if response:
                return response
            return view_func(request, *args, **kwargs)

        return wraps(view_func)(_wrapped_view)

    return decorator


def permission_required_or_403(perm, *args, **kwargs):
    """Check if user has permission, if not return 403.

    This decorator is wrapper for `permission_required` decorator.
    The only difference between `permission_required` decorator is that this
    one always set `return_403` parameter to `True`.

    Note:
        The standard Django's `permission_required` decorator redirects user to login page
        when a permission check failed.
        This decorator may be used to return HttpResponseForbidden (status 403)
        instead of redirection.
    """
    kwargs["return_403"] = True
    return permission_required(perm, *args, **kwargs)


def permission_required_or_404(perm, *args, **kwargs):
    """Check if user has permission, if not return 404.

    This decorator is wrapper for permission_required decorator.
    The only difference between `permission_required` decorator is that this
    one always set `return_404` parameter to `True`.

    Note:
        The standard Django's `permission_required` decorator redirects users to the login page
        when a permission check fails.
        This decorator may be used to return HttpResponseNotFound (status 404)
        instead of redirection.
    """
    kwargs["return_404"] = True
    return permission_required(perm, *args, **kwargs)



from django.contrib.auth import get_user_model
from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from guardian.forms import BaseObjectPermissionsForm


class BaseObjectPermissionsFormTests(TestCase):
    def setUp(self):
        self.user = get_user_model().objects.create_user("joe", "joe@example.com", "joe")
        self.obj = ContentType.objects.create(model="bar", app_label="fake-for-guardian-tests")

    def test_not_implemented(self):
        class MyUserObjectPermissionsForm(BaseObjectPermissionsForm):
            def __init__(formself, user, *args, **kwargs):
                self.user = user
                super().__init__(*args, **kwargs)

        form = MyUserObjectPermissionsForm(self.user, self.obj, {})
        self.assertRaises(NotImplementedError, form.save_obj_perms)

        field_name = form.get_obj_perms_field_name()
        self.assertTrue(form.is_valid())
        self.assertEqual(len(form.cleaned_data[field_name]), 0)



from typing import Any, Iterable, Optional

from django.contrib.auth import get_user_model
from django.db import models
from django.db.models import Model
from django.http import HttpRequest

from guardian.conf import settings
from guardian.core import ObjectPermissionChecker
from guardian.ctypes import get_content_type
from guardian.exceptions import WrongAppError


def check_object_support(obj: Model) -> bool:
    """Checks if given `obj` is supported

    Returns:
         `True` if given `obj` is supported
    """
    # Backend checks only object permissions (isinstance implies that obj
    # is not None)
    # Backend checks only permissions for Django models
    return isinstance(obj, models.Model)


def check_user_support(user_obj: Any) -> tuple[bool, Any]:
    """Checks if given user is supported.

    Checks if the given user is supported. Anonymous users need explicit
    activation via ANONYMOUS_USER_NAME

    Returns:
        A tuple of checkresult and `user_obj` which should be used for permission checks
    """
    # This is how we support anonymous users - simply try to retrieve User
    # instance and perform checks for that predefined user
    if not user_obj.is_authenticated:
        # If anonymous user permission is disabled, then they are always unauthorized
        if settings.ANONYMOUS_USER_NAME is None:
            return False, user_obj
        user_model = get_user_model()
        lookup = {user_model.USERNAME_FIELD: settings.ANONYMOUS_USER_NAME}
        user_obj = user_model.objects.get(**lookup)

    return True, user_obj


def check_support(user_obj: Any, obj: Model) -> Any:
    """Checks if given user and object are supported.

    Combination of `check_object_support` and `check_user_support`
    """
    obj_support = check_object_support(obj)
    user_support, user_obj = check_user_support(user_obj)
    return obj_support and user_support, user_obj


class ObjectPermissionBackend:
    """Django backend for checking object-level permissions."""

    supports_object_permissions = True
    supports_anonymous_user = True
    supports_inactive_user = True

    def authenticate(self, request: HttpRequest, username: Optional[str] = None, password: Optional[str] = None) -> Any:
        return None

    def has_perm(self, user_obj: Any, perm: str, obj: Optional[Model] = None) -> bool:
        """Check if a user has the permission for a given object.

        Returns `True` if given `user_obj` has `perm` for `obj`.
        If no `obj` is given, `False` is returned.
        The main difference between Django's `ModelBackend` is that we can pass
        `obj` instance here and `perm` doesn't have to contain
        `app_label` as it can be retrieved from given `obj`.

        **Inactive user support**

        If `user` is authenticated but inactive at the same time, all checks
        always return `False`.

        Note:
           Remember, that if user is not *active*, all checks would return `False`.

        Parameters:
            user_obj (User): User instance.
            perm (str): Permission string.
            obj (Model): Model instance.

        Returns:
            `True` if `user_obj` has permission, `False` otherwise.
        """

        # check if user_obj and object are supported
        support, user_obj = check_support(user_obj, obj)
        if not support:
            return False

        if "." in perm:
            app_label, _ = perm.split(".", 1)
            # TODO (David Graham): Check if obj is None or change the method signature
            if app_label != obj._meta.app_label:  # type: ignore[union-attr]
                # Check the content_type app_label when permission
                # and obj app labels don't match.
                ctype = get_content_type(obj)
                if app_label != ctype.app_label:
                    raise WrongAppError(
                        "Passed perm has app label of '%s' while "
                        "given obj has app label '%s' and given obj"
                        "content_type has app label '%s'" % (app_label, obj._meta.app_label, ctype.app_label)  # type: ignore[union-attr]
                    )

        check = ObjectPermissionChecker(user_obj)
        return check.has_perm(perm, obj)

    def get_all_permissions(self, user_obj: Any, obj: Optional[Model] = None) -> Iterable[str]:
        """Returns all permissions for a given object.

        Parameters:
            user_obj (User): User instance.
            obj (Model): Django Model instance.

        Returns:
             a set of permission strings that the given `user_obj` has for `obj`.
        """
        # check if user_obj and object are supported
        support, user_obj = check_support(user_obj, obj)
        if not support:
            return set()

        check = ObjectPermissionChecker(user_obj)
        return check.get_perms(obj)



import unittest

from django.conf import UserSettingsHolder, settings
from django.utils.functional import wraps

from guardian.conf import settings as guardian_settings


def skipUnlessTestApp(obj):
    app = "guardian.testapp"
    return unittest.skipUnless(app in settings.INSTALLED_APPS, "app %r must be installed to run this test" % app)(obj)


class TestDataMixin:
    def setUp(self):
        super().setUp()
        from django.contrib.auth import get_user_model
        from django.contrib.auth.models import Group

        User = get_user_model()
        Group.objects.create(pk=1, name="admins")
        jack_group = Group.objects.create(pk=2, name="jackGroup")
        User.objects.get_or_create(username=guardian_settings.ANONYMOUS_USER_NAME)
        jack = User.objects.create(username="jack", is_active=True, is_superuser=False, is_staff=False)
        jack.groups.add(jack_group)


class override_settings:
    """
    Acts as either a decorator, or a context manager. If it's a decorator it
    takes a function and returns a wrapped function. If it's a contextmanager
    it's used with the `with` statement. In either event entering/exiting
    are called before and after, respectively, the function/block is executed.
    """

    def __init__(self, **kwargs):
        self.options = kwargs
        self.wrapped = settings._wrapped

    def __enter__(self):
        self.enable()

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()

    def __call__(self, test_func):
        from django.test import TransactionTestCase

        if isinstance(test_func, type) and issubclass(test_func, TransactionTestCase):
            original_pre_setup = test_func._pre_setup
            original_post_teardown = test_func._post_teardown

            def _pre_setup(innerself):
                self.enable()
                original_pre_setup(innerself)

            def _post_teardown(innerself):
                original_post_teardown(innerself)
                self.disable()

            test_func._pre_setup = _pre_setup
            test_func._post_teardown = _post_teardown
            return test_func
        else:

            @wraps(test_func)
            def inner(*args, **kwargs):
                with self:
                    return test_func(*args, **kwargs)

        return inner

    def enable(self):
        override = UserSettingsHolder(settings._wrapped)
        for key, new_value in self.options.items():
            setattr(override, key, new_value)
        settings._wrapped = override

    def disable(self):
        settings._wrapped = self.wrapped
